#class Segtypes(Enum):
# Context free operations
from ..util.error import EoqSynError

# TODO implement CNS operator
# TODO implement RGX operator

class QrySegTypes:
    OBJ = 'OBJ' #object reference
    HIS = 'HIS' #history reference
    
    # Left only element-wise operations
    PTH = 'PTH' # path
    CLS = 'CLS' # class
    INO = 'INO' # instance of
    MET = 'MET' # meta
    NOT = 'NOT' # boolean not
    TRM = 'TRM' # terminate
    CNS = 'CNS' # constraint
    TRY = 'TRY' #try catch

    # Left only list-of-elements-wise operations
    IDX = 'IDX' # index
    SEL = 'SEL' # selector
    ARR = 'ARR' # outer array
    ZIP = 'ZIP' # inner array creation
    QRY = 'QRY' # cmd
    
    ANY = 'ANY' # at least one element from the right is found in the left --> Bool
    ALL = 'ALL' # all elements from the right are found in the left --> Bool
    
    # Left-vs-right element-wise operators
    EQU = 'EQU' # equal
    EQA = 'EQA' # equal any of the right elements
    NEQ = 'NEQ' # not equal
    LES = 'LES' # less
    GRE = 'GRE' # greater
    RGX = 'RGX' # regex (string only)
    
    ##generic operators
    ADD = 'ADD' # OR, addition,
    SUB = 'SUB' # XOR, subtraction,
    MUL = 'MUL' # AND, multiply,
    DIV = 'DIV' # NAND, divided,
    
    ##logic operator synonyms
    ORR = 'ORR' # OR synonym
    XOR = 'XOR' # XOR synonym
    AND = 'AND' # AND synonym
    NAD = 'NAD' # NAND synonym
    
    #Left-vs-right list-of-element-wise operators
    CSP = 'CSP' # cross product
    ITS = 'ITS' # intersection
    DIF = 'DIF' # set subtraction / relative complement
    UNI = 'UNI' # union
    CON = 'CON' # concut

#define a list of symbols for the textual representation
QRY_SYMBOLS = {
    'OBJ' : '#', 
    'HIS' : '$', 
    'PTH' : '/', 
    'CLS' : '!', 
    'INO' : '?',
    'MET' : '@',
    'NOT' : '&NOT',
    'TRM' : '&TRM',
    'CNS' : '&CNS',
    'TRY' : '&TRY',

    'IDX' : ':', 
    'SEL' : '{', 
    'ARR' : '[', 
    'ZIP' : '&ZIP',
    'QRY' : '(', 
    'ANY' : '&ANY',
    'ALL' : '&ALL', 
    'EQU' : '=',
    'EQA' : '&EQA',
    'NEQ' : '~', 
    'LES' : '<', 
    'GRE' : '>',
    'RGX' : '&RGX',
    'ADD' : '&ADD', 
    'SUB' : '&SUB', 
    'MUL' : '&MUL', 
    'DIV' : '&DIV', 
    'ORR' : '&ORR', 
    'XOR' : '&XOR', 
    'AND' : '&AND', 
    'NAD' : '&NAD', 
    'CSP' : '&CSP', 
    'ITS' : '^', 
    'DIF' : '\\',
    'UNI' : '_', 
    'CON' : '|'
}  
#free symbols: *`�;
#reserved symbols: {}()[]'"+-.,
        

# META OPERATORS 
# Use only three letters for meta operators

#element operators
class QryMetaSegTypes:
    CLS = 'CLASS' #class
    CLN = 'CLASSNAME' #class name
    CON = 'CONTAINER' #parent (container)
    PAR = 'PARENT' #parent (container)
    ALP = 'ALLPARENTS' #parent (container)
    ASO = 'ASSOCIATES' #ASSOCIATES(start=root) all elements refering to this one beginning at start. default is root
    IDX = 'INDEX' #index within its containment
    CFT = 'CONTAININGFEATURE' #the feature that contains the element
    FEA = 'FEATURES' #all features
    FEV = 'FEATUREVALUES' #all feature values
    FEN = 'FEATURENAMES' #all feature names
    ATT = 'ATTRIBUTES' #all attribute features
    ATN = 'ATTRIBUTENAMES' #all attribute feature names
    ATV = 'ATTRIBUTEVALUES' #all attribute feature values
    REF = 'REFERENCES' #all reference features
    REN = 'REFERENCENAMES' #all reference feature names
    REV = 'REFERENCEVALUES' #all reference feature values
    CNT = 'CONTAINMENTS' #all containment features
    CNV = 'CONTAINMENTVALUES' #all containment feature values
    CNN = 'CONTAINMENTNAMES' #all containment feature names
    
    #class operators
    PAC = 'PACKAGE' #class
    STY = 'SUPERTYPES' #directly inherited classes
    ALS = 'ALLSUPERTYPES' #all and also indirectly inherited classes
    IMP = 'IMPLEMENTERS' #all direct implementers of a class
    ALI = 'ALLIMPLEMENTERS' #all and also indirect implementers of a class  
    MMO = 'METAMODELS' #retrieve all metamodels
    
    #Control flow operators 
    IFF = 'IF' #if(condition,then,else);  #DEPRICATED
    TRY = 'TRY' #catch errors and return a default #NOT IMPLEMENTED
    
    
    #list operators
    LEN = 'SIZE' #size of a list #DEPRICATED
    
    #recursive operators
    REC = 'REPEAT' #REPEAT(<query>,depth) repeat a given query until no more results are found #NOT IMPLEMENTED
    
    
    
class QryIdxTypes:
    #structure operators
    FLT = 'FLATTEN' #flatten any sub list structure to a list #NOT IMPLEMENTED
    LEN = 'SIZE' #size of a list
    ASC = 'SORTASC' #sort ascending #NOT IMPLEMENTED
    DSC = 'SORTDSC' #sort descending #NOT IMPLEMENTED

        
class Seg:
    def __init__(self,stype,args):
        self.qry = stype
        self.v = args
        
    def __repr__(self):
        return QRY_SYMBOLS[self.qry]+str(self.v)
        
    def __eq__(self, other):
        if(isinstance(other, Seg)):
            return self.qry == other.qry and self.v == other.v
        else: 
            return False #can not be equal if it is a different type
        
class PthSeg(Seg):
    def __init__(self,name):
        super().__init__(QrySegTypes.PTH,name)
        
class SelSeg(Seg):
    def __init__(self,query):
        super().__init__(QrySegTypes.SEL,query)
        
    def __repr__(self):
        return '{'+str(self.v)+'}'
    
class ArrSeg(Seg):
    def __init__(self,query):
        super().__init__(QrySegTypes.ARR,query)
        
    def __repr__(self):
        return '['+str(self.v)+']'
    
class ObjSeg(Seg):
    def __init__(self,obj):
        super().__init__(QrySegTypes.OBJ,obj)
        
    def __lt__(self, other):
        return self.v < other.v
        
class Qry(Seg): #Essential object cmd 
    def __init__(self,root=None):
        if(None==root):
            super().__init__(QrySegTypes.QRY, []) #add the root to the list if it is given, otherwise add an empty list
        elif(isinstance(root,Seg)):
            super().__init__(QrySegTypes.QRY, [root])
        else:
            raise EoqSynError('If root is given it must be a query segment itself. No primitives or lists are allowed. Use Arr() to pack a list of elements.')
    def __repr__(self):
        queryStr = ''
        for seg in self.v:
            queryStr += str(seg)
        return '('+queryStr+')'
    
    def _(self,seg): #adds an existing segment
        self.v.append(seg)
    
    def Obj(self,v):
        self.v.append(ObjSeg(v))
        return self
    
    def His(self,v):
        self.v.append(Seg(QrySegTypes.HIS,v))
        return self
        
    def Pth(self,name):
        self.v.append(PthSeg(name))
        return self
    
    def Cls(self,clazz):
        self.v.append(Seg(QrySegTypes.CLS,clazz))
        return self
    
    def Ino(self,clazz):
        self.v.append(Seg(QrySegTypes.INO,clazz))
        return self
    
    def Met(self,name,args=[]):
        self.v.append(Seg(QrySegTypes.MET,[name]+args))
        return self
    
    def Not(self):
        self.v.append(Seg(QrySegTypes.NOT,None))
        return self
    
    def Trm(self,cond=None,default=None):
        self.v.append(Seg(QrySegTypes.TRM,[cond,default]))
        return self
        
    def Cns(self):
        self.v.append(Seg(QrySegTypes.CNS,None))
        return self
    
    def Try(self,query,default=None):
        self.v.append(Seg(QrySegTypes.TRY,[query,default]))
        return self
    
    def Idx(self,n):
        self.v.append(Seg(QrySegTypes.IDX,n))
        return self
    
    def Sel(self,query):
        self.v.append(SelSeg(query))
        return self
    
    def Arr(self,elements):
        self.v.append(ArrSeg(elements))
        return self

    def Zip(self,elements):
        self.v.append(Seg(QrySegTypes.ZIP,elements))
        return self
    
    def Any(self,query):
        self.v.append(Seg(QrySegTypes.ANY,query))
        return self
    
    def All(self,query):
        self.v.append(Seg(QrySegTypes.ALL,query))
        return self
    
    def Equ(self,query):
        self.v.append(Seg(QrySegTypes.EQU,query))
        return self
    
    def Eqa(self,query):
        self.v.append(Seg(QrySegTypes.EQA,query))
        return self
    
    def Neq(self,query):
        self.v.append(Seg(QrySegTypes.NEQ,query))
        return self
    
    def Les(self,query):
        self.v.append(Seg(QrySegTypes.LES,query))
        return self
    
    def Gre(self,query):
        self.v.append(Seg(QrySegTypes.GRE,query))
        return self
        
    def Rgx(self,query):
        self.v.append(Seg(QrySegTypes.RGX,query))
        return self
    
    def Rgx(self,query):
        self.v.append(Seg(QrySegTypes.RGX,query))
        return self
    
    def Add(self,query):
        self.v.append(Seg(QrySegTypes.ADD,query))
        return self
    
    def Sub(self,query):
        self.v.append(Seg(QrySegTypes.SUB,query))
        return self
    
    def Mul(self,query):
        self.v.append(Seg(QrySegTypes.MUL,query))
        return self
    
    def Div(self,query):
        self.v.append(Seg(QrySegTypes.DIV,query))
        return self
    
    def Orr(self,query):
        self.v.append(Seg(QrySegTypes.ORR,query))
        return self
    
    def Xor(self,query):
        self.v.append(Seg(QrySegTypes.XOR,query))
        return self
    
    def And(self,query):
        self.v.append(Seg(QrySegTypes.AND,query))
        return self
    
    def Nad(self,query):
        self.v.append(Seg(QrySegTypes.NAD,query))
        return self
    
    def Csp(self,query):
        self.v.append(Seg(QrySegTypes.CSP,query))
        return self
    
    def Its(self,query):
        self.v.append(Seg(QrySegTypes.ITS,query))
        return self
    
    def Dif(self,query):
        self.v.append(Seg(QrySegTypes.DIF,query))
        return self
    
    def Uni(self,query):
        self.v.append(Seg(QrySegTypes.UNI,query))
        return self
    
    def Con(self,query):
        self.v.append(Seg(QrySegTypes.CON,query))
        return self
    
''' Shortcuts to start queries '''
    
class Obj(Qry):
    def __init__(self,v):
        super().__init__()
        self.Obj(v)
        
class His(Qry):
    def __init__(self,v):
        super().__init__()
        self.His(v)

class Pth(Qry):
    def __init__(self,name):
        super().__init__()
        self.Pth(name)
    
class Cls(Qry):
    def __init__(self,clazz):
        super().__init__()
        self.Cls(clazz)
        
class Ino(Qry):
    def __init__(self,clazz):
        super().__init__()
        self.Ino(clazz)
        
class Try(Qry):
    def __init__(self,query,default):
        super().__init__()
        self.Try(query,default)
        
class Met(Qry):
    def __init__(self,name,args=[]):
        super().__init__()
        self.Met(name,args)
        
class Not(Qry):
    def __init__(self):
        super().__init__()
        self.Not()
        
class Cns(Qry):
    def __init__(self):
        super().__init__()
        self.Cns()
        
class Idx(Qry):
    def __init__(self,n):
        super().__init__()
        self.Idx(n)
        
class Sel(Qry):
    def __init__(self,query):
        super().__init__()
        self.Sel(query)
        
class Arr(Qry):
    def __init__(self,elements):
        super().__init__()
        self.Arr(elements)       
    
class Zip(Qry):
    def __init__(self,elements):
        super().__init__()
        self.Zip(elements)   
        
class Any(Qry):
    def __init__(self,select):
        super().__init__()
        self.Any(select)
        
class All(Qry):
    def __init__(self,select):
        super().__init__()
        self.All(select)
        
class Equ(Qry):
    def __init__(self,query):
        super().__init__()
        self.Equ(query)
        
class Eqa(Qry):
    def __init__(self,query):
        super().__init__()
        self.Eqa(query)
        
class Neq(Qry):
    def __init__(self,query):
        super().__init__()
        self.Neq(query)
        
class Les(Qry):
    def __init__(self,query):
        super().__init__()
        self.Les(query)
        
class Gre(Qry):
    def __init__(self,query):
        super().__init__()
        self.Gre(query)
        
class Rgx(Qry):
    def __init__(self,query):
        super().__init__()
        self.Rgx(query)        

def QrySegFactory(qry,v):
    if(QrySegTypes.OBJ==qry):
        return ObjSeg(v)
    else:
        return Seg(qry,v)

        
        

        
        



'''
 2019 Bjoern Annighoefer
'''

from .domain import Domain
from ..util.logger import NoLogging,LogLevels
from ..serialization import TextSerializer
from ..command.command import Get
from ..event import ALL_EVENT_TYPES

class CmdRunnerBasedDomain(Domain):
    def __init__(self,cmdRunner,logger=NoLogging(),serializer=TextSerializer(),enableBenchmark=False):
        super().__init__(logger)
        self.serializer = serializer
        self.enableBenchmark = enableBenchmark
        self.doCounter = 0
        self.cmdRunner = cmdRunner
#         if(self.cmdRunner):
#             self.cmdRunner.Observe(self._OnCmdRunnerEvents)
        
    def RawDo(self,cmd,sessionId=None):
        self.doCounter += 1
        self.logger.PassivatableLog(LogLevels.INFO,lambda : "cmd: %s"%(self.serializer.serialize(cmd)))
        res = self.cmdRunner.Exec(cmd,sessionId)
        return res
        
    def Get(self,target):
        cmd = Get(target)
        res = self.Do(cmd)
        return res
    
    #Override the event provid methods since the sole event provider shall be the cmd runner
    #@Override
    def Observe(self,callback,eventTypes=ALL_EVENT_TYPES,context=None,sessionId=None): #by default register for all events
        self.cmdRunner.Observe(callback,eventTypes,context,sessionId)
    
    #@Override    
    def Unobserve(self,callback,context=None):
        self.cmdRunner.Unobserve(callback,context)
        
    #@Override
    def NotifyObservers(self,evts,excludedCallback=None,excludedContext=None):
        self.cmdRunner.NotifyObservers(evts,excludedCallback,excludedContext)
        
    def Close(self):
        #benchmark:
        if self.enableBenchmark:
            self.cmdRunner.benchmark.SaveToFile('CmdBenchmark.csv')
            self.logger.Info("Command benchmark saved to CmdBenchmark.csv")
            self.cmdRunner.qryEvaluator.benchmark.SaveToFile('QryBenchmark.csv')
            self.logger.Info("Query segment benchmark saved to QryBenchmark.csv")
            
    


'''
 Bjoern Annighoefer 2019
'''

from ..event import EvtProvider

'''
 ACCESSORS
'''    

#Defintion of Accessor base class
class MdbAccessor(EvtProvider): 
    def Lock(self):
        pass
    
    def Release(self):
        pass
    
    ''' METAMODEL ACESSORS '''
   
    def GetAllMetamodels(self):
        pass
     
    def RegisterMetamodel(self,package):
        pass
    
    def UnregisterMetamodel(self,package):
        pass

    ''' MODEL ACESSORS '''
   
    def GetRoot(self):
        pass
    
    def Get(self,obj,feature):
        pass
    
    def Set(self,obj,feature,value):
        pass
    
    def GetParent(self,obj):
        pass
    
    def GetAllParents(self,obj):
        pass
    
    def GetAssociates(self,obj):
        pass
    
    def GetIndex(self,obj):
        pass
    
    def GetContainingFeature(self,obj):
        pass
        
    def Add(self,obj,feature,child):
        pass
        
    def Remove(self,obj,feature,child):
        pass
    
    def Move(self,obj,newIndex):
        pass

    def Clone(self,obj,mode):
        pass
    
    def Create(self,clazz,n,constructorArgs=[]):
        pass
    
    def CreateByName(self,packageName,className,n,constructorArgs=[]):
        pass
    
    def GetClassByName(self,packageName,className):
        pass
    
    def Class(self,obj):
        pass
    
    def ClassName(self,obj):
        pass
    
    ''' CLASS ACESSORS '''
   
    def Package(self,clazz):
        pass
    
    def Supertypes(self,clazz):
        pass
    
    def AllSupertypes(self,clazz):
        pass
    
    def Implementers(self,clazz):
        pass
    
    def AllImplementers(self,clazz):
        pass
    
    def GetAllChildrenOfType(self,obj,clazz):
        pass
    
    def GetAllChildrenInstanceOfClass(self,obj,clazz):
        pass
    
    def GetAllFeatures(self,obj):
        pass
    
    def GetAllFeatureNames(self,obj):
        pass
    
    def GetAllFeatureValues(self,obj):
        pass
    
    def GetAllAttributes(self,obj):
        pass
    
    def GetAllAttributeNames(self,obj):
        pass
    
    def GetAllAttributeValues(self,obj):
        pass
    
    def GetAllReferences(self,obj):
        pass
    
    def GetAllReferenceNames(self,obj):
        pass
    
    def GetAllReferenceValues(self,obj):
        pass
    
    def GetAllContainments(self,obj):
        pass
    
    def GetAllContainmentNames(self,obj):
        pass
    
    def GetAllContainmentValues(self,obj):
        pass
    
    
    

    



'''
 Bjoern Annighoefer 2019
 '''

from .framehandler import FrameHandler
from .frame import FrameTypes
from ..event import ALL_EVENT_TYPES

class HandlerInfo:
    def __init__(self,handler,version):
        self.version = version
        self.handler = handler
    
class MultiVersionFrameHandler(FrameHandler):
    def __init__(self,handlerInfos=[]):
        super().__init__()
        self.handlers = []
        for handlerInfo in handlerInfos:
            self.AddHandler(handlerInfo[0], handlerInfo[1]) #handler infos must be list of tuples
        
    def Handle(self,frames,sessionId=None):
        for i in range(len(frames)):
            wasHandled = False
            for handlerInfo in self.handlers:
                if(frames[i].ver >= handlerInfo.version):
                    frames[i] = handlerInfo.handler.Handle([frames[i]],sessionId)[0]
                    wasHandled = True
                    break
            if(not wasHandled):
                errorMsg = "Could not find a frame handler for version %d"%(frames[i].ver)
                print("WARNING: %s"%(errorMsg))
                frames[i].eoq = FrameTypes.ERR
                frames[i].dat = errorMsg
        return frames   
    
    #@Override 
    def Gby(self,sessionId):
        for handlerInfo in self.handlers:
            handler = handlerInfo.handler
            handler.Gby(sessionId)
        
    def AddHandler(self,version,handler):
        self.handlers.append(HandlerInfo(handler,version))
        self.handlers = sorted(self.handlers, key=lambda a: a.version,reverse=True)
        #handler.ObserveEvents(self.OnHandlerEvent)
        
    #@Override 
    def Observe(self,callback,eventTypes=ALL_EVENT_TYPES,context=None,sessionId=None): #by default register for all events
        for handlerInfo in self.handlers:
            handler = handlerInfo.handler
            handler.Observe(callback,eventTypes,context,sessionId)
    
    #@Override    
    def Unobserve(self,callback,context=None):
        for handlerInfo in self.handlers:
            handler = handlerInfo.handler
            handler.Unobserve(callback,context)
        


'''
 Command result
 
'''
from ..util.error import EoqError
from .command import CmdTypes

class ResTypes:
    OKY = 'OKY'
    ERR = 'ERR'

class Res:
    def __init__(self,commandType,status,value,transactionId=0,changeId=0):
        self.res = commandType #correspondes to the command that this result comes from
        self.s = status #status
        self.v = value #result of the executed command (error text in case of error, subresult array in case of a compound command)
        self.n = transactionId
        self.c = changeId
        
    def GetValue(self):
        return ResGetValue(self)
    
    
def ResGetValue(res):
    val = None
    if(ResTypes.ERR==res.s):
        raise EoqError(0,res.v)
    if(CmdTypes.CMP==res.res):
        val = [sr.v for sr in res.v]
    else: 
        val = res.v
    return val


'''
 Bjoern Annighoefer 2019
'''

from .serializer import Serializer

'''

JSON Serializer

'''
   
from ..frame.frame import Frame
from ..command.command import Cmd
from ..query.query import QrySegFactory
from ..command.result import Res
from ..event.event import Evt

import json

def JsonSerializeHook(obj):
    return obj.__dict__

def JsonDeserializeHook(dct):
        if "eoq" in dct:
            return Frame(dct["eoq"],dct["uid"],dct["dat"],version=dct["ver"])
        if "cmd" in dct:
            return Cmd(dct["cmd"],dct["a"])
        elif "qry" in dct: 
            return QrySegFactory(dct["qry"],dct["v"])
        elif "res" in dct:
            return Res(dct["res"],dct["s"],dct["v"],dct["n"],dct["c"])
        elif "evt" in dct:
            return Evt(dct["evt"],dct["k"],dct["a"])
        return dct


class JsonSerializer(Serializer):
    def __init__(self):
        pass

    def Ser(self, val):
        try:
            return json.dumps(val,default=JsonSerializeHook)
        except Exception as e:
            return "{error: '%s'"%(str(e))
    
    def Des(self,code):
        return json.loads(code,object_hook=JsonDeserializeHook)


from pyecore.resources import global_registry
from .constraintmodel import getEClassifier, eClassifiers
from .constraintmodel import name, nsURI, nsPrefix, eClass
from .constraintmodel import Constraint, ConstraintTypeEnum


from . import constraintmodel

__all__ = ['Constraint', 'ConstraintTypeEnum']

eSubpackages = []
eSuperPackage = None
constraintmodel.eSubpackages = eSubpackages
constraintmodel.eSuperPackage = eSuperPackage


otherClassifiers = [ConstraintTypeEnum]

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [constraintmodel] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



'''
	Author Bjoern Annighoefer Mar 2019
'''

from .error import EoqParseError
from .model import *

### CONSTANTS

#context constants
CONTEXT_TYPE_DONTCARE = 1 #Don't check the class type
CONTEXT_TYPE_EXPLICIT = 2 #Check the class type

#return multiplicity

CONTEXT_MULTIPLICITY_UNCHANGED = 1 #always return a list
CONTEXT_MULTIPLICITY_FLATTENED = 2 #always return a list of depth 1
CONTEXT_MULTIPLICITY_FORCESINGLE = 3 #always return a single object

#segment type constants
SEGTYPE_PATH = 2
SEGTYPE_CLASS = 3
SEGTYPE_INSTANCE = 4
SEGTYPE_ID = 5
SEGTYPE_META = 6
SEGTYPE_RESOURCE = 7
SEGTYPE_HISTORY = 8
SEGTYPE_LISTOP = 9

#selector types
SELTYPE_EQUAL = 1 #=
SELTYPE_DIFFERENT = 2 #~
SELTYPE_GREATER = 3 #>
SELTYPE_SMALLER = 4 #<

#selector value types
VALTYPE_INT = 1
VALTYPE_BOOL = 2
VALTYPE_FLOAT = 3
VALTYPE_STRING = 4
VALTYPE_ID = 5
VALTYPE_HISTORY = 6
VALTYPE_UNSET = 7

INDTYPE_NUMBER = 1 #1,2,3,4,5,6
INDTYPE_RANGE = 2 #1:3,:
INDTYPE_APPEND = 3 #+ (add after the last element. Only valid for add operations)
INDTYPE_REMOVE = 4 #- (remove element from array or unset attribute or reference)

################ MAIN CLASS ###############

class QueryParser:
	@staticmethod
	def __ResolveSegment(queryStr,segInfo):
		identifier = queryStr[segInfo['identifierStart']:segInfo['identifierEnd']+1]
		
		segment = None
		if(SEGTYPE_PATH==segInfo['type']):
			segment = PathSegment()
		elif(SEGTYPE_CLASS==segInfo['type']):
			segment = ClassSegment()
		elif(SEGTYPE_INSTANCE==segInfo['type']):
			segment = InstanceOfSegment()
		elif(SEGTYPE_ID==segInfo['type']):
			segment = IdSegment()
		elif(SEGTYPE_HISTORY==segInfo['type']):
			segment = HistorySegment()
		elif(SEGTYPE_META==segInfo['type']):
			if(identifier in MetaSegmentIdentifiersE):
				segment = MetaSegment()
				segment.identifierType = MetaSegmentIdentifiersE.from_string(identifier)
			else: 
				raise EoqParseError('ERROR(META_IDENTIFIER)',"Unknown meta segment: %s"%(identifier),queryStr)
		elif(SEGTYPE_LISTOP==segInfo['type']):
			if(identifier in ListOpSegmentTypesE):
				segment = ListOpSegment()
				segment.identifierType = ListOpSegmentTypesE.from_string(identifier)
			else: 
				raise EoqParseError('ERROR(LISTOP_IDENTIFIER)',"Unknown list operation: %s"%(identifier),queryStr)
		elif(SEGTYPE_RESOURCE==segInfo['type']):
			segment = ResourceSegment()
		else:
			raise EoqParseError('ERROR(SEGMENT_TYPE)',"Unknown segment type %d"%(segInfo['type']),queryStr)

		segment.identifier = identifier
		
		#check for correct identifier values
		
		if(segInfo['hasSelector']):
			segment.selector = Selector()
			segment.selector.name = queryStr[segInfo['selectorStart']:segInfo['selectorEnd']+1]
			## Get the value an comperator
			valueStr = queryStr[segInfo['valueStart']:segInfo['valueEnd']+1]
			if(VALTYPE_INT == segInfo['valueType'] or VALTYPE_FLOAT == segInfo['valueType'] or VALTYPE_STRING == segInfo['valueType']):
				if(SELTYPE_EQUAL == segInfo['selectorType']):
					segment.selector.operator = EqualOperator()
				elif(SELTYPE_DIFFERENT == segInfo['selectorType']):
					segment.selector.operator = NotEqualOperator()
				elif(SELTYPE_GREATER == segInfo['selectorType']):
					segment.selector.operator = GreaterOperator()
				elif(SELTYPE_SMALLER == segInfo['selectorType']):
					segment.selector.operator = LessOperator()
				else:
					raise EoqParseError("ERROR: Unknown comperator %d"%(segInfo['selectorType']),queryStr)
				#set the selector
				if(VALTYPE_INT == segInfo['valueType']):
					segment.selector.value = IntValue(v=int(valueStr))
				elif(VALTYPE_FLOAT == segInfo['valueType']):
					segment.selector.value = FloatValue(v=float(valueStr))
				elif(VALTYPE_STRING == segInfo['valueType']):
					segment.selector.value = StringValue(v=valueStr)
				
			elif(VALTYPE_BOOL == segInfo['valueType'] or VALTYPE_ID == segInfo['valueType'] or VALTYPE_HISTORY == segInfo['valueType'] or VALTYPE_UNSET == segInfo['valueType']):
				if(SELTYPE_EQUAL == segInfo['selectorType']):
					segment.selector.operator = EqualOperator()
				elif(SELTYPE_DIFFERENT == segInfo['selectorType']):
					segment.selector.operator = NotEqualOperator()
				elif(SELTYPE_GREATER == segInfo['selectorType']):
					raise EoqParseError(1,"Unsupported comperator > (%d)"%(segInfo['selectorType']),queryStr)
				elif(SELTYPE_SMALLER == segInfo['selectorType']):
					raise EoqParseError(2,"Unsupported comperator < (%d)"%(segInfo['selectorType']),queryStr)
				
				if(VALTYPE_BOOL == segInfo['valueType']):
					segment.selector.value = BoolValue()
					if('true' == valueStr):
						segment.selector.value.v = True
					elif('false' == valueStr):
						segment.selector.value.v = False
					else:
						raise EoqParseError(3,"Unsupported boolean value '%s'. Should be 'true' or 'false'."%(valueStr),queryStr)
					
				elif(VALTYPE_ID == segInfo['valueType']):
					segment.selector.value = ObjectRefValue(v=int(valueStr))
					
				elif(VALTYPE_HISTORY == segInfo['valueType']):
					segment.selector.value = HistoryRefValue(v=int(valueStr))
				
				elif(VALTYPE_UNSET == segInfo['valueType']):
					segment.selector.value = EmptyValue()
				
		#evaluate depth
		#depth = 1 #default depth for segments
		if(segInfo['hasDepth']):
			segment.depth = Depth()
			segment.depth.value = int(queryStr[segInfo['depthStart']:segInfo['depthEnd']+1])
# 		else:
# 			if(SEGTYPE_CLASS==segInfo['type']) or (SEGTYPE_INSTANCE==segInfo['type']):
# 				segment.depth = Depth()
# 				segment.depth.value = 1 #default depth for class recovery is infinity
			
		#consider the index
		if(segInfo['hasIndex']):
			indexStr = queryStr[segInfo['indexStart']:segInfo['indexEnd']+1]
			if(INDTYPE_NUMBER==segInfo['indexType']):
				segment.index = NumberIndex()
				segment.index.value = int(indexStr)
			elif(INDTYPE_RANGE==segInfo['indexType']):
				lowerUpper = indexStr.split(':')
				segment.index = RangeIndex()
				if(2 == len(lowerUpper)):
					segment.index.lower = int(lowerUpper[0])
					segment.index.upper = int(lowerUpper[1])
				elif(1==len(lowerUpper) and ':' == indexStr[0]):
					segment.index.lower = 1
					segment.index.upper = int(lowerUpper[0])
				elif(1==len(lowerUpper) and ':' == indexStr[-1]):
					segment.index.lower = int(lowerUpper[0])
					segment.index.upper = 0 #last element
				else:
					raise EoqParseError(4,"Bad index string '%s'. Possible is [2],[2:4],[+],[-]"%(indexStr),queryStr)
			elif(INDTYPE_APPEND==segInfo['indexType']):
				segment.index = AddIndex()
			elif(INDTYPE_REMOVE==segInfo['indexType']):
				segment.index = RemoveIndex()
			else:
				#Should never go here
				raise EoqParseError(5,"ERROR: Unsupported index type %d."%(segInfo['indexType']),queryStr)
		return segment
	
	@staticmethod
	def __RetrieveSourceClass(queryStr,contextInfo):
		sourceClass = None
		sourceClassName = queryStr[contextInfo['nameStart']:contextInfo['nameEnd']+1]
		if(not '*' == sourceClassName):
			sourceClass = SourceClass()
			sourceClass.name = sourceClassName
			sourceClass.dontCare = False
		return sourceClass
	
	@staticmethod
	def __RetrieveReturnMultiplicity(queryStr,contextInfo):
		returnMultiplicity = None
		if(contextInfo['contextMultiplicity'] == CONTEXT_MULTIPLICITY_FORCESINGLE):
			returnMultiplicity = ForceSingleReturnMultiplicity()
		elif(contextInfo['contextMultiplicity'] == CONTEXT_MULTIPLICITY_UNCHANGED):
			returnMultiplicity = UnchangedReturnMuliplicity()
		elif(contextInfo['contextMultiplicity'] == CONTEXT_MULTIPLICITY_FLATTENED):
			returnMultiplicity = FlattenedReturnMultiplicity()
		return returnMultiplicity

	@staticmethod
	def StringToQuery(queryStr):
		query = Query()
		
		contextInfo = {}
		segInfo = {}
		
		#definition of states for queryStr parser
		#state constants
		START = 1
		CONTEXT_CLASS_START = 2
		CONTEXT_CLASS = 3
		
		CONTEXT_MULTIPLICITY_START = 4
		CONTEXT_MULTIPLICITY = 5
		
		PATH_START = 6
		IDENTIFIER_START = 7
		IDENTIFIER = 8
		DEPTH_START = 9
		DEPTH = 10
		SELECTOR_START = 11
		SELECTOR = 12
		VALUE_START = 13
		VALUE = 14
		INDEX_START = 15
		INDEX = 16
		
		state = START
		
		#query context information
		contextInfo['contextType'] = CONTEXT_TYPE_DONTCARE
		contextInfo['contextClass'] = None
		contextInfo['contextMultiplicity'] = CONTEXT_MULTIPLICITY_UNCHANGED
		
		contextInfo['nameStart'] = 0
		contextInfo['nameEnd'] = 0
		
		#properties of a segment
		segInfo['type'] = 0
		segInfo['hasDepth'] = False
		segInfo['hasSelector'] = False
		segInfo['hasIndex'] = False
		
		#variables for parsing
		segInfo['identifierStart'] = 0
		segInfo['identifierEnd'] = 0
		segInfo['depthStart'] = 0
		segInfo['depthEnd'] = 0
		segInfo['selectorType'] = 0
		segInfo['selectorStart'] = 0
		segInfo['selectorEnd'] = 0
		segInfo['valueType'] = 0
		segInfo['valueStart'] = 0
		segInfo['valueEnd'] = 0
		segInfo['indexType'] = 0
		segInfo['indexStart'] = 0
		segInfo['indexEnd'] = 0
		
		n = len(queryStr)
		
		for i in range(n):
			c = queryStr[i]
			if(START == state):
				if('(' == c):
					contextInfo['nameStart'] = i+1
					state = CONTEXT_CLASS_START
				elif('/' == c):
					segInfo['type'] = SEGTYPE_PATH
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('$' == c):
					segInfo['type'] = SEGTYPE_CLASS
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('#' == c):
					segInfo['type'] = SEGTYPE_ID
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('?' == c):
					segInfo['type'] = SEGTYPE_INSTANCE
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('@' == c):
					segInfo['type'] = SEGTYPE_META
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif(':' == c):
					segInfo['type'] = SEGTYPE_RESOURCE
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('!' == c):
					segInfo['type'] = SEGTYPE_HISTORY
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('.' == c):
					segInfo['type'] = SEGTYPE_LISTOP
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				else:
					raise EoqParseError(10,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(CONTEXT_CLASS_START == state):
				if(c.isalpha() or '*'==c):
					contextInfo['nameEnd'] = i
					state = CONTEXT_CLASS
				else:
					raise EoqParseError(11,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(CONTEXT_CLASS == state):
				if(c.isalpha() or c.isdigit() or '*'==c):
					contextInfo['nameEnd'] = i
					state = CONTEXT_CLASS
				elif(':'==c):
					#contextInfo = self.__EvaluateContext(root,queryStr,contextInfo)
					sourceClass = QueryParser.__RetrieveSourceClass(queryStr,contextInfo)
					query.sourceClass = sourceClass
					state = CONTEXT_MULTIPLICITY_START
				elif(')'==c):
					#contextInfo['contextMultiplicity'] = CONTEXT_MULTIPLICITY_UNCHANGED
					sourceClass = QueryParser.__RetrieveSourceClass(queryStr,contextInfo)
					query.sourceClass = sourceClass
					state = PATH_START
				else:
					raise EoqParseError(12,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(CONTEXT_MULTIPLICITY_START == state):
				if('1'==c):
					contextInfo['contextMultiplicity'] = CONTEXT_MULTIPLICITY_FORCESINGLE
					state = CONTEXT_MULTIPLICITY
				elif('_'==c):
					contextInfo['contextMultiplicity'] = CONTEXT_MULTIPLICITY_FLATTENED
					state = CONTEXT_MULTIPLICITY
				elif('*'==c):
					contextInfo['contextMultiplicity'] = CONTEXT_MULTIPLICITY_UNCHANGED
					state = CONTEXT_MULTIPLICITY
				else:
					raise EoqParseError(13,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(CONTEXT_MULTIPLICITY == state):
				if(')'==c):
					returnMultiplicity = QueryParser.__RetrieveReturnMultiplicity(queryStr,contextInfo)
					query.returnMultiplicity = returnMultiplicity
					state = PATH_START
				else:
					raise EoqParseError(14,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(PATH_START == state):
				if('/' == c):
					segInfo['type'] = SEGTYPE_PATH
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('$' == c):
					segInfo['type'] = SEGTYPE_CLASS
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('#' == c):
					segInfo['type'] = SEGTYPE_ID
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('?' == c):
					segInfo['type'] = SEGTYPE_INSTANCE
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('@' == c):
					segInfo['type'] = SEGTYPE_META
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif(':' == c):
					segInfo['type'] = SEGTYPE_RESOURCE
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('!' == c):
					segInfo['type'] = SEGTYPE_HISTORY
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('.' == c):
					segInfo['type'] = SEGTYPE_LISTOP
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				else:
					raise EoqParseError(15,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			
			elif(IDENTIFIER_START == state):
				#reset segment properties
				segInfo['hasDepth'] = False
				segInfo['hasSelector'] = False
				segInfo['hasIndex'] = False
				if(SEGTYPE_RESOURCE == segInfo['type']):
					if(c.isalpha() or '_' == c or '.' == c):
						segInfo['identifierEnd'] = i
						state = IDENTIFIER
					else:
						raise EoqParseError(16,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
				else:
					if(c.isalpha() or '_' == c):
						segInfo['identifierEnd'] = i
						state = IDENTIFIER
					else:
						raise EoqParseError(16,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(IDENTIFIER == state):
				if(c.isalpha() or '_' == c  or ('0' <= c and c <= '9') or '\\' == c):
					segInfo['identifierEnd'] = i
				elif(SEGTYPE_RESOURCE == segInfo['type'] and '.'==c): #special case of reverse path to container
					segInfo['identifierEnd'] = i
				elif('{'==c):
					segInfo['hasSelector'] = True
					segInfo['selectorStart'] = i+1
					state = SELECTOR_START
				elif('['==c):
					segInfo['hasIndex'] = True
					segInfo['indexStart'] = i+1
					state = INDEX_START
				elif('<'==c):
					segInfo['hasDepth'] = True
					segInfo['depthStart'] = i+1
					state = DEPTH_START
				elif('/' == c):
					segment = QueryParser.__ResolveSegment(queryStr,segInfo)
					query.segments.add(segment)
					segInfo['type'] = SEGTYPE_PATH
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('$' == c):
					segment = QueryParser.__ResolveSegment(queryStr,segInfo)
					query.segments.add(segment)
					segInfo['type'] = SEGTYPE_CLASS
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('?' == c):
					segment = QueryParser.__ResolveSegment(queryStr,segInfo)
					query.segments.add(segment)
					segInfo['type'] = SEGTYPE_INSTANCE
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('#' == c):
					segment = QueryParser.__ResolveSegment(queryStr,segInfo)
					query.segments.add(segment)
					segInfo['type'] = SEGTYPE_ID
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('@' == c):
					segment = QueryParser.__ResolveSegment(queryStr,segInfo)
					query.segments.add(segment)
					segInfo['type'] = SEGTYPE_META
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('!' == c):
					segment = QueryParser.__ResolveSegment(queryStr,segInfo)
					query.segments.add(segment)
					segInfo['type'] = SEGTYPE_HISTORY
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
				elif('.' == c):
					segment = QueryParser.__ResolveSegment(queryStr,segInfo)
					query.segments.add(segment)
					segInfo['type'] = SEGTYPE_LISTOP
					segInfo['identifierStart'] = i+1
					state = IDENTIFIER_START
# 				elif(':' == c): # There can not be a second resource
# 					segInfo['type'] = SEGTYPE_RESOURCE
# 					segInfo['identifierStart'] = i+1
# 					state = IDENTIFIER_START
				else:
					raise EoqParseError(17,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(DEPTH_START == state):
				if(('0' <= c and c <= '9') or '-' == c):
					segInfo['depthEnd'] = i
					state = DEPTH
				else:
					raise EoqParseError(18,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(DEPTH == state):
				if('0' <= c and c <= '9'):
					segInfo['depthEnd'] = i
				elif('>'==c):
					state = IDENTIFIER
				else:
					raise EoqParseError(19,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(SELECTOR_START == state):
				if(c.isalpha()):
					segInfo['selectorEnd'] = i
					state = SELECTOR
				else:
					raise EoqParseError(20,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(SELECTOR == state):
				if(c.isalpha()):
					segInfo['selectorEnd'] = i
				elif('='==c):
					segInfo['selectorType'] = SELTYPE_EQUAL
					segInfo['valueStart'] = i+1
					state = VALUE_START
				elif('~'==c):
					segInfo['selectorType'] = SELTYPE_DIFFERENT
					segInfo['valueStart'] = i+1
					state = VALUE_START
				elif('>'==c):
					segInfo['selectorType'] = SELTYPE_GREATER
					segInfo['valueStart'] = i+1
					state = VALUE_START
				elif('<'==c):
					segInfo['selectorType'] = SELTYPE_SMALLER
					segInfo['valueStart'] = i+1
					state = VALUE_START
				else:
					raise EoqParseError(22,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(VALUE_START == state):
				if(('0' <= c and c <= '9') or '-' == c):
					segInfo['valueEnd'] = i
					segInfo['valueType'] = VALTYPE_INT
					state = VALUE
				elif('.'==c):
					segInfo['valueEnd'] = i
					segInfo['valueType'] = VALTYPE_FLOAT
					state = VALUE
				elif('t'==c or 'f'==c):
					segInfo['valueEnd'] = i
					segInfo['valueType'] = VALTYPE_BOOL
					state = VALUE
				elif('#'==c):
					segInfo['valueStart'] = i+1
					segInfo['valueEnd'] = i+1
					segInfo['valueType'] = VALTYPE_ID
					state = VALUE
				elif('$'==c):
					segInfo['valueStart'] = i+1
					segInfo['valueEnd'] = i+1
					segInfo['valueType'] = VALTYPE_HISTORY
					state = VALUE
				elif('\''==c):
					segInfo['valueStart'] = i+1
					segInfo['valueEnd'] = i+1
					segInfo['valueType'] = VALTYPE_STRING
					state = VALUE
				elif('%'==c):
					segInfo['valueStart'] = i
					segInfo['valueEnd'] = i
					segInfo['valueType'] = VALTYPE_UNSET
					state = VALUE
				else:
					raise EoqParseError(23,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(VALUE == state):
				if('}'==c):
					state = IDENTIFIER
				elif(VALTYPE_INT==segInfo['valueType'] and ('0' <= c and c <= '9')):
					segInfo['valueEnd'] = i
				elif(VALTYPE_ID==segInfo['valueType'] and ('0' <= c and c <= '9')):
					segInfo['valueEnd'] = i
				elif(VALTYPE_HISTORY==segInfo['valueType'] and ('0' <= c and c <= '9')):
					segInfo['valueEnd'] = i
				elif(VALTYPE_INT==segInfo['valueType'] and'.'==c):
					segInfo['valueEnd'] = i
					segInfo['valueType'] = VALTYPE_FLOAT
				elif(VALTYPE_BOOL==segInfo['valueType'] and c.isalpha()):
					segInfo['valueEnd'] = i
				elif(VALTYPE_STRING==segInfo['valueType'] and '\''!=c):
					segInfo['valueEnd'] = i
				elif(VALTYPE_STRING==segInfo['valueType'] and '\''==c):
					segInfo['valueEnd'] = i-1 # not necessary, but a statement is required
				else:
					raise EoqParseError(24,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(INDEX_START == state):
				if('0' <= c and c <= '9'):
					segInfo['indexEnd'] = i
					segInfo['indexType'] = INDTYPE_NUMBER
					state = INDEX
				elif('+' == c):
					segInfo['indexEnd'] = i
					segInfo['indexType'] = INDTYPE_APPEND
					state = INDEX
				elif('-' == c):
					segInfo['indexEnd'] = i
					segInfo['indexType'] = INDTYPE_REMOVE
					state = INDEX
				elif(':' == c):
					segInfo['indexEnd'] = i
					segInfo['indexType'] = INDTYPE_RANGE
					state = INDEX
				else:
					raise EoqParseError(25,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
			elif(INDEX == state):
				if('0' <= c and c <= '9') and (INDTYPE_NUMBER == segInfo['indexType'] or INDTYPE_RANGE == segInfo['indexType']):
					segInfo['indexEnd'] = i
				elif(':' == c) and (INDTYPE_NUMBER == segInfo['indexType'] or INDTYPE_RANGE == segInfo['indexType']):
					segInfo['indexEnd'] = i
					segInfo['indexType'] = INDTYPE_RANGE
					state = INDEX
				elif(']'==c):
					state = IDENTIFIER
				else:
					raise EoqParseError(26,'Unexpected character \'%c\' at %d'%(c,i),queryStr)
		
		#evaluate last segment
		segment = QueryParser.__ResolveSegment(queryStr,segInfo)
		query.segments.add(segment)
		#result = lastSegmentHandler(root,queryStr,segInfo)
		return query
	
	@staticmethod
	def QueryToString(query):
		queryStr = ''
		if(query.sourceClass or query.returnMultiplicity):
			sourceClassName = ('*' if(not query.sourceClass or query.sourceClass.dontCare) else query.sourceClass.name)
			multiplicityChar = ('*' if(not query.returnMultiplicity) else query.returnMultiplicity.symbol)
			queryStr += '(%s:%c)'%(sourceClassName,multiplicityChar)
		for segment in query.segments:
			queryStr += '%c%s'%(segment.startCharacter,segment.identifier)
			if(segment.selector):
				selector = segment.selector
				value = selector.value
				valueStr = ''
				if(value.type == ValueTypesE.INT):
					valueStr = str(value.v)
				elif(value.type == ValueTypesE.FLOAT):
					valueStr = str(value.v)
				elif(value.type == ValueTypesE.BOOL):
					valueStr = ('true' if value.v else 'false')
				elif(value.type == ValueTypesE.STRING):
					valueStr = '\'%s\''%(value.v)
				elif(value.type == ValueTypesE.OBJECTREF):
					valueStr = '#%d'%(value.v)
				elif(value.type == ValueTypesE.HISTORYREF):
					valueStr = '$%d'%(value.v)
				elif(value.type == ValueTypesE.EMPTY):
					valueStr = '#'
				queryStr += '{%s%c%s}'%(selector.name,selector.operator.symbol,valueStr)
			if(segment.index):
				index = segment.index
				if(index.type == IndexTypesE.NUMBER):
					queryStr += '[%d]'%(index.value)
				elif(index.type == IndexTypesE.RANGE):
					queryStr += '[%d:%d]'%(index.lower,index.upper)
				elif(index.type == IndexTypesE.ADD):
					queryStr += '[+]'
				elif(index.type == IndexTypesE.REMOVE):
					queryStr += '[-]'
			if(segment.depth):
				queryStr += '<%d>'%(segment.depth.value)
		return queryStr


#check for all actions in folder

from .. import CommandParser,ResultParser,ValueParser
from ..model import *
from ..error import EoqActionError
from ..domains import LocalDomain
from ..domains import MultiprocessingQueueDomainHost
from ..domains import MultiprocessingQueueDomainClient
from . import ActionUtils

import sys
import os
import glob
import imp
import time
import traceback

import io
from contextlib import redirect_stdout
import datetime

from multiprocessing import Process
import threading
from .actionutils import ActionUtils
from ..domains import LocalDomain
from paramiko import channel
from sys import stderr

class WriteRedirector:
    def __init__(self,callId,channelName,domain,legacyWriteFcn=None,bufferTime=1.0):
        self.callId = callId
        self.domain = domain
        self.channelName = channelName
        self.legacyWriteFcn = legacyWriteFcn
        self.bufferTime = bufferTime
        
        self.lastWriteTime = time.perf_counter()
        self.buffer = b''

    def write(self,data, *args):
            currentWriteTime = time.perf_counter()
            try:
                if(self.legacyWriteFcn):
                    self.legacyWriteFcn(data)
                if(isinstance(data, str) and 0<len(data)):
                    data = data.encode('utf-8')
                    self.buffer+=data
                    if(currentWriteTime-self.lastWriteTime>self.bufferTime):
                        ActionUtils.OutputToChannel(self.callId,self.channelName,self.buffer,self.domain)
                        self.buffer = b''
                        self.lastWriteTime = currentWriteTime
            except:
                pass
    def flush(self):
            try:
                if(0<len(self.buffer)):
                    ActionUtils.OutputToChannel(self.callId,self.channelName,self.buffer,self.domain)
                    self.buffer = b''
            except:
                pass
        
    def getStdStreamCompatibleWriteRedirector(self):
        return (lambda data, *args: self.write(data, *args))


def AsyncCallProcessMain(actiondir,callId,actionName,argStr,cmdQueue,resQueue):
    #connect to domain
    domain = MultiprocessingQueueDomainClient(cmdQueue,resQueue)
    stdoutRedirector = WriteRedirector(callId,'STDOUT',domain,legacyWriteFcn=sys.stdout.write)
    stderrRedirector = WriteRedirector(callId,'STDERR',domain,legacyWriteFcn=sys.stderr.write)
    #redirect std outputs
    sys.stdout.write = stdoutRedirector.getStdStreamCompatibleWriteRedirector()  
    sys.stderr.write = stderrRedirector.getStdStreamCompatibleWriteRedirector()
    #load action function
    searchString = os.path.join(actiondir+'/*.py')
    pythonfiles = glob.glob(searchString)
    actionFunction = None
    for pythonfile in pythonfiles:
        pathname, filename = os.path.split(pythonfile)
        actionname = filename[:-3]; #without .py
        if(actionName == actionname):
            packagename = pathname.replace('/','.')
            modulename = packagename+'.'+actionname
            file, modulepath, description = imp.find_module(actionname, [pathname])
            actionmodule = imp.load_module(modulename, file, modulepath, description)
            actionFunction = getattr(actionmodule,actionname)
            break
    #convert input values
    if(actionFunction):
        args = ValueParser.StringToValue(argStr)
        pArguments = ValueParser.ValueToPython(args)
        ActionUtils.SetCallStatus(callId,CallStatusE.RUNNING,domain)
        pResult = actionFunction(domain,*pArguments)
        returnValues = ValueParser.PythonToValue(pResult)
        #make sure all output was printed to the domain
        stdoutRedirector.flush()
        stderrRedirector.flush()
        #finaly change the call status to FINISHED
        ActionUtils.SetCallStatus(callId,CallStatusE.FINISHED,domain)

#PRIVATE CLASS DEFINITIONS
class EoqExternalActionError(EoqActionError):
    def __init__(self, code, message):
        EoqActionError.__init__(self, code, "EXTERNALACTIONHANDLER: "+message)
        
class ActionInfo:
    def __init__(self,name,details,actionFunction,args,results,description):
        self.name = name
        self.details = details
        self.actionFunction = actionFunction
        self.args = args
        self.results = results
        self.description = description
        
class ArgumentInfo:
    def __init__(self,name,type,min,max,description,default,choices):
        self.name = name
        self.type = type
        self.min = min
        self.max = max
        self.description = description
        self.default = default
        self.choices = choices
        
class RunningAction:
    def __init__(self,actionCall,process,domainWrapper):
        self.actionCall = actionCall
        self.process = process
        self.domainWrapper = domainWrapper
        
class RunAsCurrentTransactionLocalDomainWrapper(LocalDomainA):
    def __init__(self,localdomain : LocalDomain, transaction : Transaction):
        self._domain = localdomain
        self.transaction = transaction
        
    def Do(self, command : CommandA) -> ResultA:
        historyOffset = len(self.transaction.history)
        self.__AdoptHistoryRefsCommand(command,historyOffset)
        result = self._domain.DoAsTransaction(command,self.transaction)
        return result
            
    def Call(self, name : str, args : ListValue, transaction : Transaction) -> CallResult:
        return LocalDomainA.Call(self, name, args, transaction)
    
    def __AdoptHistoryRefsCommand(self,command,offset):
        if(command.type == CommandTypesE.COMPOUND):
            for command in command.commands: 
                self.__AdoptHistoryRefsCommand(command,offset)
        elif(command.type == CommandTypesE.RETRIEVE):
            self.__AdoptHistoryRefsValue(command.target,offset)
        elif(command.type == CommandTypesE.UPDATE):
            self.__AdoptHistoryRefsValue(command.target,offset)
            self.__AdoptHistoryRefsValue(command.value,offset)
        elif(command.type == CommandTypesE.CALL):
            self.__AdoptHistoryRefsValue(command.args,offset)
        elif(command.type == CommandTypesE.ASYNCCALL):
            self.__AdoptHistoryRefsValue(command.args,offset)
        return command

    def __AdoptHistoryRefsValue(self,value,offset):
        if(value.type == ValueTypesE.LIST):
            for v in value.v:
                self.__AdoptHistoryRefsValue(v,offset)
        elif(value.type == ValueTypesE.HISTORYREF):
            value.v += offset
        return value
        
'''
    The external action handler
'''
class ExternalActionHandler(ActionHandlerA):
    
    def __init__(self,domain : LocalDomainA, basedir='actions'):
        self._domain = domain
        self._basedir = basedir
        self._actions = {}
        self._runningActions = {}
        #register default action for reloading action files        
        self.__ReloadActions(domain)
        return
    
    def __ScanForActions(self):
        #find all actions within files
        searchString = os.path.join(self._basedir+'/*.py')
        pythonfiles = glob.glob(searchString)
        sys.path.append(self._basedir) #add this to make libraries of actions accessible
        for pythonfile in pythonfiles:
            pathname, filename = os.path.split(pythonfile)
            actionname = filename[:-3]; #without .py
            packagename = pathname.replace('/','.')
            modulename = packagename+'.'+actionname
            try: #prevent server crash by broken action scripts
                file, modulepath, description = imp.find_module(actionname, [pathname])
                actionmodule = imp.load_module(modulename, file, modulepath, description)
                #check for a unique name
                if(actionname in self._actions):
                    print("WARNING: Found external action %s in file %s, but an action with the same name is already registered from file %s. Action names must be unique!"%(actionname,pythonfile,self._actions[actionname].filepath))
                    continue
                #check if the module contains a function with the same name
                if(actionname not in dir(actionmodule)):
                    print("WARNING: Found no function called %s in %s. This action can not be used."%(actionname,pythonfile))
                    continue
                print("INFO: Found external action %s in file %s"%(actionname,pythonfile))
                #register the new action 
                
                actionFunction = getattr(actionmodule,actionname)
                actionArguments = self.__ParseActionArguments(actionname,actionFunction)
                actionResults = self.__ParseActionResults(actionname,actionFunction)
                actionDescription = actionmodule.__doc__
                self._actions[actionname] = ActionInfo(actionname,pythonfile,actionFunction,actionArguments,actionResults,actionDescription)
            except Exception as e:
                print("Error loading external action '%s':"%(pythonfile))
                traceback.print_exc(file=sys.stdout)
            
    def __ParseActionArguments(self,actionName,actionFunction):
            args = []
            nArguments = actionFunction.__code__.co_argcount
            functionVariables = actionFunction.__code__.co_varnames
            actionArguments = functionVariables[0:nArguments]
            actionArgumentTypeInfos = actionFunction.__annotations__
            
            if(nArguments==0):
                print('WARNING: Action %s has no argument. Actions must have at least one argument of type EoqDomainA.'%(actionName))
            else:
                #look for the first argument. This must always be a domain
                argumentName = actionArguments[0]
                argument = None
                if(argumentName not in actionArgumentTypeInfos):
                    print('WARNING: Argument 1 of action %s is not annotated assuming EoqDomainA as type.'%(actionName))
                else:
                    argumentAnnotation = actionArgumentTypeInfos[argumentName]
                    if(DomainA.__name__!=argumentAnnotation):
                        print('WARNING: Argument 1 of action %s has type %s but expected is %s. This will probably not work.'%(actionName,argumentAnnotation,DomainA.__name__))
                #add all args after the first 
                for i in range(1,nArguments):
                    argumentName = actionArguments[i]
                    argument = None
                    if(argumentName not in actionArgumentTypeInfos):
                        print('WARNING: Argument %s of action %s has no annotation. Assuming * as type'%(argumentName,actionName))
                        argument = ArgumentInfo(argumentName,'*',1,1,'','',[])
                    else:
                        argumentAnnotation = actionArgumentTypeInfos[argumentName]
                        [argumentType, argumentMin, argumentMax, description, default, choices] = self.__ParseArgumentAnnotation(argumentAnnotation)
                        argument = ArgumentInfo(argumentName,argumentType, argumentMin, argumentMax, description, default, choices)
                    args.append(argument)
            return args
        
    def __ParseActionResults(self,actionName,actionFunction):
            args = []
            nArguments = actionFunction.__code__.co_argcount
            functionVariables = actionFunction.__code__.co_varnames
            actionArguments = functionVariables[0:nArguments]
            actionArgumentTypeInfos = actionFunction.__annotations__
            
            argumentName = 'return' #'return' is the default key for return annotations in python
            argument = None
            if(argumentName not in actionArgumentTypeInfos):
                print('WARNING: Action %s has no return annotation. Assuming no return value'%(actionName))
            else:
                argumentAnnotation = actionArgumentTypeInfos[argumentName]
                if('' != argumentAnnotation):
                    [argumentType, argumentMin, argumentMax, description, default, choices] = self.__ParseArgumentAnnotation(argumentAnnotation)
                    args.append(ArgumentInfo(argumentName,argumentType, argumentMin, argumentMax, description, default, choices))
            return args
        
    def __ParseArgumentAnnotation(self,argumentAnnotation):
        #a annotation should look like this: <Type>[<multiplicity=0..1|*>]{<choice1>',<choice2>,...}=<default>:<Description>
        #all parameters except type are optional
        argumentType = ''
        argumentMin = 1
        argumentMax = 1
        description = ''
        default = ''
        choices = []
        
        #parse string
        nArgumentAnnotation = len(argumentAnnotation)+1
        multiplicityStart = argumentAnnotation.find('[')
        choiceStart = argumentAnnotation.find('{')
        defaultStart = argumentAnnotation.find('=')
        descriptionStart = argumentAnnotation.find(':')
        
        #multiplicity
        if(multiplicityStart > 0):
            multiplicityEnd = multiplicityStart+argumentAnnotation[multiplicityStart:].find(']')
            if(multiplicityEnd>0 and multiplicityEnd>multiplicityStart):
                multiplicityStr = argumentAnnotation[multiplicityStart+1:multiplicityEnd]
                if('..' in multiplicityStr):
                    multiplicity = multiplicityStr.split('..')
                    if(len(multiplicity)==2):
                        argumentMin = int(multiplicity[0])
                        if('*' == multiplicity[1]):
                            argumentMax = -1
                        else:
                            argumentMax = int(multiplicity[1])
                else:
                    if('*' == multiplicityStr):
                        argumentMin = 0
                        argumentMax = -1
                    else:
                        argumentMin = int(multiplicityStr)
                        argumentMax = int(multiplicityStr)
            else:
                print('WARNING: Malformed multiplicity in argument annotation %s'%(argumentAnnotation))

        #choice
        if(choiceStart > 0):
            choiceEnd = choiceStart+argumentAnnotation[choiceStart:].find('}')
            if(choiceEnd>0 and choiceEnd-1>choiceStart+1): #skip { and }
                choiceStr = argumentAnnotation[choiceStart+1:choiceEnd]
                choices = choiceStr.split(',')
            else:
                print('WARNING: Malformed choice in argument annotation %s'%(argumentAnnotation))
        #default
        if(defaultStart > 0):
            defaultEnd = min([nArgumentAnnotation,
                              descriptionStart%nArgumentAnnotation])
            if(defaultEnd>0 and defaultEnd-1>defaultStart+1):
                default = argumentAnnotation[defaultStart+1:defaultEnd] #skip :' and '
            else:
                print('WARNING: Malformed default in argument annotation %s'%(argumentAnnotation))

        #description
        if(descriptionStart > 0):
            descriptionEnd = nArgumentAnnotation 
            if(descriptionEnd>0 and descriptionEnd>descriptionStart+1): #skip :
                description = argumentAnnotation[descriptionStart+1:descriptionEnd] #skip :
            else:
                print('WARNING: Malformed description in argument annotation %s'%(argumentAnnotation))


        #type
        typeStart = 0
        typeEnd = min([nArgumentAnnotation,
                       multiplicityStart%nArgumentAnnotation,
                       choiceStart%nArgumentAnnotation,
                       defaultStart%nArgumentAnnotation,
                       descriptionStart%nArgumentAnnotation])
        argumentType = argumentAnnotation[typeStart:typeEnd]                 
        return [argumentType, argumentMin, argumentMax, description, default, choices] 
        
        
    def __AddActionToDomain(self,action,domain):
        actionDescriptor = None
        #check if an action with that name already exists
        command = CommandParser.RetrieveCommandStr(ObjectRefValue(v=0),"/actions{name='%s'}"%(action.name))
        result = domain.Do(command)
        if(ResultParser.IsResultNok(result)):
            raise EoqExternalActionError(1,'Could not read actions from domain')
        if(result.value.type == ValueTypesE.LIST and len(result.value.v) == 1 and result.value.v[0].type == ValueTypesE.OBJECTREF):
            actionDescriptor = result.value.v[0]
            
        #update or add an action
        if(actionDescriptor):
            command = CompoundCommand()
            command.commands.add(self.__ActionUpdateCommand(action,actionDescriptor))                   
            domain.Do(command)
        else:
            command = CompoundCommand()
            command.commands.add(CommandParser.CreateCommand('http://www.eoq.de/model/v1.0', 'ActionInfo',1))
            command.commands.add(self.__ActionUpdateCommand(action,HistoryRefValue(v=0)))        
            command.commands.add(CommandParser.UpdateCommandStr(ObjectRefValue(v=0), '/actions[+]', HistoryRefValue(v=0)))                          
            result = domain.Do(command)
            if(ResultParser.IsResultNok(result)):
                raise EoqExternalActionError(2,'Could not add action %s to domain because: %s(%d)%s'%(action.name,result.type,result.code,result.message))
            actionDescriptor = result.results[0].value
        parameterDescriptors = []
        for parameter in action.args:
            parameterDescriptors.append(self.__AddParameterToAction(actionDescriptor, parameter, 'parameters',domain))
        #delete existing parameters
        command = CommandParser.RetrieveCommandStr(actionDescriptor,"/parameters")
        result = domain.Do(command)
        actualParameters = result.value
        for actualParameter in actualParameters.v:
            if(actualParameter.v not in [x.v for x in parameterDescriptors]):
                command = CommandParser.UpdateCommandStr(actionDescriptor, '/parameters[-]', actualParameter)                          
                result = domain.Do(command)
        
        actionDescriptors = []
        for parameter in action.results:
            actionDescriptors.append(self.__AddParameterToAction(actionDescriptor, parameter, 'results',domain))
            
        #delete existing results
        command = CommandParser.RetrieveCommandStr(actionDescriptor,"/results")
        result = domain.Do(command)
        actualParameters = result.value
        for actualParameter in actualParameters.v:
            if(actualParameter.v not in [x.v for x in actionDescriptors]):
                command = CommandParser.UpdateCommandStr(actionDescriptor, '/results[-]', actualParameter)                          
                result = domain.Do(command)
                
        return actionDescriptor
    
    def __AddParameterToAction(self,actionDescriptor,parameter,target,domain):
        parameterDescriptor = None
        #check if an parameter with that name already exists
        command = CommandParser.RetrieveCommandStr(actionDescriptor,"/%s{name='%s'}"%(target,parameter.name))
        result = domain.Do(command)
        if(ResultParser.IsResultNok(result)):
            raise EoqExternalActionError(1,'Could not read %s for action'%(target))
        if(result.value.type == ValueTypesE.LIST and len(result.value.v) == 1 and result.value.v[0].type == ValueTypesE.OBJECTREF):
            parameterDescriptor = result.value.v[0]
        if(parameterDescriptor):
            command = CompoundCommand()
            command.commands.add(self.__ActionParameterUpdateCommand(parameter,parameterDescriptor))                   
            domain.Do(command)
        else:
            command = CompoundCommand()
            command.commands.add(CommandParser.CreateCommand('http://www.eoq.de/model/v1.0', 'ActionParameter',1))
            command.commands.add(self.__ActionParameterUpdateCommand(parameter,HistoryRefValue(v=0))) 
            command.commands.add(CommandParser.UpdateCommandStr(actionDescriptor, '/%s[+]'%(target), HistoryRefValue(v=0)))                          
            result = domain.Do(command)
            if(ResultParser.IsResultNok(result)):
                raise EoqExternalActionError(2,'Could not add %s %s to domain because: %s(%d)%s'%(target,parameterDescriptor.name,result.type,result.code,result.message))
            parameterDescriptor = result.results[0].value
        choiceDescriptors = []
        for choice in parameter.choices:
            choiceDescriptors.append(self.__AddChoiceToParameter(parameterDescriptor, choice,domain))
        #delete existing ones
        command = CommandParser.RetrieveCommandStr(parameterDescriptor,"/choices")
        result = domain.Do(command)
        actualChoices = result.value
        for actualChoice in actualChoices.v:
            if(actualChoice.v not in [x.v for x in choiceDescriptors]):
                command = CommandParser.UpdateCommandStr(parameterDescriptor, '/choices[-]', actualChoice)                          
                result = domain.Do(command)
                
        return parameterDescriptor
    
    def __AddChoiceToParameter(self,parameterDescriptor,choice,domain):
        choiceDescriptor = None
        #check if an parameter with that name already exists
        command = CommandParser.RetrieveCommandStr(parameterDescriptor,"/choices{value='%s'}"%(choice))
        result = domain.Do(command)
        if(ResultParser.IsResultNok(result)):
            raise EoqExternalActionError(1,'Could not read parameters for action')
        if(result.value.type == ValueTypesE.LIST and len(result.value.v) == 1 and result.value.v[0].type == ValueTypesE.OBJECTREF):
            choiceDescriptor = result.value.v[0]
        if(choiceDescriptor):
            pass
        else:
            command = CompoundCommand()
            command.commands.add(CommandParser.CreateCommand('http://www.eoq.de/model/v1.0', 'Choice',1))
            command.commands.add(CommandParser.UpdateCommandStr(HistoryRefValue(v=0), '/value', StringValue(v=choice))) 
            command.commands.add(CommandParser.UpdateCommandStr(parameterDescriptor, '/choices[+]', HistoryRefValue(v=0)))                         
            result = domain.Do(command)
            if(ResultParser.IsResultNok(result)):
                raise EoqExternalActionError(2,'Could not add choice %s to parameter because: %s(%d)%s'%(choice,result.type,result.code,result.message))
            choiceDescriptor = result.results[0].value
        return choiceDescriptor
    
    def __ActionUpdateCommand(self,action,actionDescriptor):
        command = CompoundCommand()
        command.commands.add(CommandParser.UpdateCommandStr(actionDescriptor, '/name', StringValue(v=action.name)))
        command.commands.add(CommandParser.UpdateCommandStr(actionDescriptor, '/details', StringValue(v=action.details)))
        command.commands.add(CommandParser.UpdateCommandStr(actionDescriptor, '/description', StringValue(v=action.description)))
        command.commands.add(CommandParser.UpdateCommandStr(actionDescriptor, '/handler', ObjectValue(v=self)))
        return command
    
    def __ActionParameterUpdateCommand(self,parameter,parameterDescriptor):
        command = CompoundCommand()
        command.commands.add(CommandParser.UpdateCommandStr(parameterDescriptor, '/name', StringValue(v=parameter.name)))
        command.commands.add(CommandParser.UpdateCommandStr(parameterDescriptor, '/type', StringValue(v=parameter.type)))
        command.commands.add(CommandParser.UpdateCommandStr(parameterDescriptor, '/lowerBound', IntValue(v=parameter.min)))
        command.commands.add(CommandParser.UpdateCommandStr(parameterDescriptor, '/upperBound', IntValue(v=parameter.max)))
        command.commands.add(CommandParser.UpdateCommandStr(parameterDescriptor, '/default', StringValue(v=parameter.default)))
        command.commands.add(CommandParser.UpdateCommandStr(parameterDescriptor, '/description', StringValue(v=parameter.description)))
        return command
    
    def __RemoveActionFromDomain(self,actionname,domain):
        actionDescriptor = self._actions[actionname]

        #update or add an action
        command = CommandParser.UpdateCommandStr(ObjectRefValue(v=0), '/actions[-]', actionDescriptor)                       
        result = domain.Do(command)
        if(ResultParser.IsResultNok(result)):
            raise EoqExternalActionError(3,'Could not remove action %s from domain because: %s(%d)%s'%(actionname,result.type,result.code,result.message))
        return
    
    def __ReloadActions(self,domain):
        #empty the list of actions
        currentlyRegisteredActions = dict(self._actions) #make a copy to delete them later
        self._actions.clear()
        #add an internal action to reload all actions
        actionname = 'reload-actions'
        actionFunction = lambda d: self.__ReloadActions(d)
        self._actions[actionname] = ActionInfo(actionname,'',actionFunction,[],[],'Reloads the action files and their descriptions from the file system')
        
        self.__ScanForActions()
        #register actions in the model
        for newActionname in self._actions:
            action = self._actions[newActionname]
            actionObject = self.__AddActionToDomain(action,domain)
                     
        #delete non existing actions from the model
        for oldActionname in currentlyRegisteredActions:
            if(oldActionname not in self._actions):
                self.__RemoveActionFromDomain(oldActionname,domain)
        return
        
    def Call(self, actionCall : ActionCall, transaction : Transaction):
        if(actionCall.action not in self._actions):
            raise EoqExternalActionError(1,'Action with name %s is unknown.'%(actionCall.action))
        action = self._actions[actionCall.action]
        #print("Called external action %s(%s)"%(name,ValueParser.ValueToString(args)))
        domain = RunAsCurrentTransactionLocalDomainWrapper(self._domain,transaction)
        pArguments = ValueParser.ValueToPython(actionCall.args)
        
        stdoutOutput = io.StringIO()
        ActionFunction = action.actionFunction
        with redirect_stdout(stdoutOutput):
            ActionUtils.SetCallStatus(actionCall.callId,CallStatusE.RUNNING,domain)
            pResult = ActionFunction(domain,*pArguments)
        returnValues = ValueParser.PythonToValue(pResult)
        if(stdoutOutput.getvalue()):
            ActionUtils.OutputToChannel(actionCall.callId,'STDOUT',stdoutOutput.getvalue(),domain)
        actionCall.returnValues = returnValues
        ActionUtils.SetCallStatus(actionCall.callId,CallStatusE.FINISHED,domain)
        return 
    
    def AsyncCall(self, actionCall : ActionCall, transaction : Transaction):
        if(actionCall.action not in self._actions):
            raise EoqExternalActionError(1,'Action with name %s is unknown.'%(actionCall.action))
        observerThread = threading.Thread(target=self.__AsyncCallObserverThread, args=(actionCall,))
        observerThread.start()
        return 
    
    def CallStatus(self, actionCall : ActionCall, transaction : Transaction):
        raise EoqExternalActionError(45,'CallStatus: Not implemented for ExternalActionHandler.')

    def AbortCall(self, actionCall : ActionCall, transaction : Transaction):
        if(actionCall.callId not in self._runningActions):
            raise EoqExternalActionError(46,'ExternalActionHandler: CallId %d is unknown'%(actionCall.callId))
        #remove from the list of running processes
        runningAction = self._runningActions.pop(actionCall.callId)
        #kill the process
        runningAction.process.terminate() 
        #set an aborted status
        domain = RunAsCurrentTransactionLocalDomainWrapper(self._domain,transaction)
        ActionUtils.SetCallStatus(actionCall.callId,CallStatusE.ABORTED,domain)
        return
    
    def __AsyncCallObserverThread(self,actionCall : ActionCall):
        argsStr = ValueParser.ValueToString(actionCall.args)
        domainWrapper = MultiprocessingQueueDomainHost(self._domain)
        proc = Process(target=AsyncCallProcessMain, args=(self._basedir,actionCall.callId,actionCall.action,argsStr,domainWrapper.cmdQueue,domainWrapper.resQueue))
        runningAction = RunningAction(actionCall,proc,domainWrapper)
        self._runningActions[actionCall.callId] = runningAction
        
        domainWrapper.Start()
        proc.start()
        
        while(proc.is_alive()): #wait until the thread has finished
            time.sleep(0.1)
        domainWrapper.Join()


'''
    A remote WebSocket domain for EOQ2
    based on the jseoq web socket domain by Bjoern Annighoefer (https://gitlab.com/eoq/js/eoq2)
    (C) 2021 Matthias Brunner
'''

import websocket
import threading
import time
import uuid
import traceback

from eoq2 import Hel,Gby
from eoq2.domain import Domain
from eoq2 import __version__ as eoqVersion
from eoq2.frame import Frame,FrameTypes


class Timeout(Exception):
    def __init__(self):
        super().__init__(self)
    
class WebSocketDomainNotConnected(Exception):
    def __init__(self):
        super().__init__()


class PendingCommand():
    def __init__(self,sessionId,cmd,timeout):
        #Session
        self.sessionId=sessionId

        # Command
        self.cmd=cmd
        self.timeout=timeout
        self.sent=time.time()

        # Results
        self.__rxEvent=threading.Event()
        self.ready=True
        self.received=None
        self.res=None

    def wait(self,timeout=None):
        return self.__rxEvent.wait(timeout)

    def setResults(self,res):
        self.read=True
        self.received=time.time()
        self.res=res
        self.__rxEvent.set()


class WebSocketDomain(Domain):
    def __init__(self,url,timeout,serializer,retries=50,retryTimeout=5,retryWait=1):
        super().__init__()

        # Retry configuration
        self.__retries=retries
        self.__retryWait=retryWait # in seconds
        self.__retryTimeout=retryTimeout # in seconds

        self.__url=url # user selected url
        self.__timeout=timeout # timeout of the connection
        self.__serializer=serializer
        self.__pendingCmds=dict()

        self.__numericVersion=int(eoqVersion.replace(".", ""))

        # To be initialized later
        self.__webSocketReady=None
        self.__webSocket=None
        self.__webSocketUrl=None
        self.__webSocketThread=None
        self.__sessionId=None

        # Open web-socket to target url
        self.__Open(self.__url)

    def __IsOpened(self):
        res=False
        if not self.__webSocket is None:
            res=True
        return res

    def __thread(self):
        try:
            self.__webSocket.run_forever()
        except KeyboardInterrupt:
            print('Websocket thread is stopping')

    def __TryOpen(self,url):
        self.__webSocketReady=threading.Event()
        self.__webSocketUrl=url      
        self.__webSocket = websocket.WebSocketApp(self.__webSocketUrl,    
                            on_open=lambda ws: self.__OnOpen(ws),                           
                            on_message=lambda ws,msg: self.__OnMessage(ws,msg),
                            on_error=lambda ws,error: self.__OnError(ws,error),
                            on_ping=lambda ws: self.__OnPing(ws),
                            on_close=lambda ws,close_code,close_reason: self.__OnClose(ws,close_code,close_reason))
        self.__webSocketThread=threading.Thread(target=self.__thread)
        self.__webSocketThread.daemon = True
        self.__webSocketThread.start()

    def __Open(self,url):
        if not self.__IsOpened():
 
            self.__TryOpen(url) # Initial try      
            ready=False
            retry=0      
                  
            while not ready and self.__retries>0:
                 self.__retries-=1                 
                 self.__TryOpen(url)
                 ready=self.__webSocketReady.wait(self.__retryTimeout)    # wait for websocket to become ready
                 retry+=1

            if ready:   # Succeeded to connect       
                self.__InitSession()
                print('Connected to EOQ remote domain at '+str(url))
            else:       # Failed to connect
                print('Failed to connect to EOQ remote domain at '+str(url))     
                raise WebSocketDomainNotConnected()    
        
        else:
            raise ValueError('Could not open WebSocket to domain: already open')
        return True
   
    def __Abort(self,reason='unknown'):
        print('Aborted, reason='+str(reason))
        self.__Close()


    def __InitSession(self):
        res=self.RawDo(Hel('user','password'))     
        self.__sessionId=res.v
        print('Session ID is : '+str(self.__sessionId))

    def __terminateSession(self):
        if self.__sessionId is not None:
            res=self.RawDo(Gby(self.__sessionId))
            print('Session with ID '+str(self.__sessionId)+' terminated')

    def __getCmdId(self):        
        return str(uuid.uuid4())       

    def RawDo(self,cmd,sessionId=None,wait=True):
        res=None
        if self.__IsOpened():
            cmdId=self.__getCmdId()
            # Register as pending command
            self.__pendingCmds[cmdId]=PendingCommand(sessionId,cmd,self.__timeout)
            frame=Frame(FrameTypes.CMD,cmdId,cmd,self.__numericVersion)
            frames=[frame]
            serializedFrames = self.__serializer.serialize(frames)
            self.__webSocket.send(serializedFrames)
        else:
            raise ValueError('Cannot process command on web-socket domain. The web-socket connection is not open.')

        response=self.__WaitForResponse(self.__timeout,cmdId)  
        if response[0]:
            res=response[1]
        else:
            raise Timeout()

        return res

    def __WaitForResponse(self,timeout,cmdId):
        res=False,None
        if cmdId in self.__pendingCmds:
            pendingCmd=self.__pendingCmds[cmdId]        
            received=pendingCmd.wait()              
            if received:
                res=True,pendingCmd.res    
        else:
            raise ValueError('Command with '+str(cmdId)+' is unknown')
        return res        

    def __OnOpen(self,ws,a=None,b=None):
        self.__webSocketReady.set()

    def __OnPing(self,ws,a=None,b=None):
        pass

    def __OnMessage(self,ws=None,serializedFrames=[]):  
        frames=self.__serializer.deserialize(serializedFrames)
        for frame in frames:            
            if frame.uid in self.__pendingCmds:
                self.__pendingCmds[frame.uid].setResults(frame.dat)
                del self.__pendingCmds[frame.uid]

    def __OnError(self,ws,error):

        ws.keep_running=False
        errorHandlers={
             'ConnectionRefusedError':lambda error: print(error)
        }

        # Shutting down WebSocket Thread
        self.__webSocket.keep_running=False       

        if error.__class__.__name__ in errorHandlers:
            errorHandlers[error.__class__.__name__](error)
        else:  
            self.__Abort('Unhandled '+error.__class__.__name__)

       

    def __OnClose(self,ws,close_code,close_reason):
        self.__sessionId=None
    
    def __Close(self):
        self.__terminateSession()
        self.__webSocket.close()
        self.__webSocket=None

    def Close(self):
        self.__Close()





'''
 Bjoern Annighoefer 2019
'''
from ..valuecodec import ValueCodec
from ...query.query import ObjSeg
from ...util.error import EoqError

from pyecore.ecore import EObject,EEnumLiteral

class PyEcoreIdCodec(ValueCodec):
    def __init__(self):
        self.objectToIdLUT = {}
        self.idToObjectLUT = {}
        self.lastId = 0
        pass
    
    def Enc(self,v):
        #v = self.__ResolveProxyAndBuildInClasses(v)
        
        if(isinstance(v,EEnumLiteral)):
            return str(v)
        elif(isinstance(v,EObject)):
            idNb = 0
            try:
                idNb = self.objectToIdLUT[v]
            except:
                idNb = self.lastId
                self.objectToIdLUT[v] = idNb
                self.idToObjectLUT[idNb] = v
                self.lastId += 1 #root id starts now at 0 like all other indexes
            return ObjSeg(idNb)
        else:
            return v
    def Dec(self,c):
        if(isinstance(c,ObjSeg)):
            obj = None
            idNb = c.v
            try:
                obj = self.idToObjectLUT[idNb]
            except:
                raise EoqError(0,'Error id #%s is not a known object'%(idNb))
            return obj
        else:
            return c
        
    '''
     PRIVATE methods
    '''
       
#     def __ResolveProxyAndBuildInClasses(self,obj): 
#         if(isinstance(obj, EProxy)):
#             obj.force_resolve()
#             obj = obj._wrapped #remove the outer proxy
#         if(isinstance(obj, (MetaEClass,type,types.ModuleType))): #this is necessary to mask compiled model instances
#             obj = obj.eClass
#         return obj


'''
 2019 Bjoern Annighoefer
'''


from ...action import CallManager
from ...frame import FrameTypes,Frame
from ...event.event import CstEvt,OupEvt,CvaEvt

class MutliprocessingQueueCallManager(CallManager):
    def __init__(self,cmdQueue,serializer,callId=-1):
        super().__init__(callId)
        self.cmdQueue = cmdQueue
        self.serializer = serializer
    
    def ChangeCallStatus(self,callId,status,info=''):
        evt = CstEvt(callId,status,info)
        frame = Frame(FrameTypes.EVT,0,evt)
        frameStr = self.serializer.Ser([frame])
        self.cmdQueue.put(frameStr)
            
    def SetCallValue(self,callId,value):
        evt = CvaEvt(callId,value)
        frame = Frame(FrameTypes.EVT,0,evt)
        frameStr = self.serializer.Ser([frame])
        self.cmdQueue.put(frameStr)
            
    def AddCallOutputRaw(self,callId,channelName,data):
        evt = OupEvt(callId,channelName,data)
        frame = Frame(FrameTypes.EVT,0,evt)
        frameStr = self.serializer.Ser([frame])
        self.cmdQueue.put(frameStr)





'''
 Bjoern Annighoefer 2019
'''

class Serializer:
    def Ser(self,val):
        pass
    def Des(self,code):
        pass
    def serialize(self,val):
        return self.Ser(val) #legacy support
    def deserialize(self,code):
        return self.Des(code) #legacy support
    
    
    

    


'''
 2019 Bjoern Annighoefer
'''

from ..event.event import EvtProvider
from ..util.logger import NoLogging
from ..command.command import Gby

class Domain(EvtProvider):
    def __init__(self,logger=NoLogging()):
        super().__init__()
        self.logger = logger
        
    def RawDo(self,cmd,sessionId=None):
        pass
    
    def Do(self,cmd,sessionId=None):
        res = self.RawDo(cmd,sessionId)
        value = res.GetValue()
        return value
    
    def Gby(self,sessionId):
        return self.Do(Gby(sessionId))
    
    def Close(self):
        pass #is to be called if the domain closes down
    
                    
    
        
        


"""
Bjoern Annighoefer Mar 2019
"""

from .error import EoqError,EoqParseError,EoqRetrieveError,EoqCreateError,EoqIdError,EoqValueError,EoqUpdateError,EoqContextError,EoqTransactionError,EoqActionError,EoqGeneralError

__version__ = "1.0.0"



'''
2019 Bjoern Annighoefer
'''

class LogLevels:
    DEBUG = "debug"
    INFO = "info"
    WARN = "warn"
    ERROR = "error"
    
DEFAULT_LOG_LEVELS = [LogLevels.INFO,LogLevels.WARN,LogLevels.ERROR]

class Logger:
    def __init__(self,activeLevels=DEFAULT_LOG_LEVELS):
        self.activeLevels = activeLevels
        
    def ShallLog(self):
        return True
    
    def Log(self,level,msg):
        if(level in self.activeLevels and self.ShallLog()):
            self._Log(level,msg)
            
    def PassivatableLog(self,level,func):
        if(level in self.activeLevels and self.ShallLog()):
            self._Log(level,func())
    
    def Debug(self,msg):
        self.Log(LogLevels.DEBUG, msg)
    
    def Info(self,msg):
        self.Log(LogLevels.INFO,msg)
        
    def Warn(self,msg):
        self.Log(LogLevels.WARN,msg)
        
    def Error(self,msg):
        self.Log(LogLevels.ERROR,msg)

    #the following must be overwritten to produce the output
    def _Log(self,level,msg):
        pass
        
        
'''
A default logger which does nothing
'''

class NoLogging(Logger):
    def __init__(self):
        super().__init__()
        
    #@Override
    def ShallLog(self):
        return False
    
    
'''
A default logger which does nothing
'''

class ConsoleLogger(Logger):
    def __init__(self,activeLevels=DEFAULT_LOG_LEVELS):
        super().__init__(activeLevels)

        
    #@Override         
    def _Log(self,level,msg):
        print("%s: %s"%(level,msg))

'''
A default logger that outputs every thing to the console and dedicated files for each active log level
'''

import os
import logging

class ConsoleAndFileLogger(Logger):
    def __init__(self,toConsole=True,toFile=True,logDir='./log',activeLevels=DEFAULT_LOG_LEVELS):
        super().__init__(activeLevels)
        self.toConsole = toConsole
        self.toFile = toFile
        self.logDir = logDir
        
        self.loggers = {}
        
        if(self.toFile):
            #make sure the
            if(not os.path.isdir(self.logDir)):
                os.makedirs(self.logDir)
            #create native python loggers for each level
            for level in self.activeLevels:
                #init error logger
                logger = logging.getLogger(level)
                logFile = os.path.join(self.logDir,"%s.log"%(level))
                fh = logging.FileHandler(logFile,'w')
                fh.setLevel(logging.INFO)
                formatter = logging.Formatter('%(asctime)s - %(message)s')
                fh.setFormatter(formatter)
                logger.addHandler(fh)
                logger.setLevel(logging.INFO)
                self.loggers[level] = logger
                
    #@Override
    def ShallLog(self):
        return (self.toConsole or self.toFile)
                
    #@Override         
    def _Log(self,level,msg):
        if(self.toConsole):
            print("%s: %s"%(level,msg))
        if(self.toFile):
            self.loggers[level].info(msg)

class FileLogger(Logger):
    def __init__(self,toConsole=False,toFile=True,logDir='./log',activeLevels=DEFAULT_LOG_LEVELS):
        super().__init__(activeLevels)
        self.toConsole = toConsole
        self.toFile = toFile
        self.logDir = logDir
        
        self.loggers = {}
        
        if(self.toFile):
            #make sure the
            if(not os.path.isdir(self.logDir)):
                os.makedirs(self.logDir)
            #create native python loggers for each level
            for level in self.activeLevels:
                #init error logger
                logger = logging.getLogger(level)
                logFile = os.path.join(self.logDir,"%s.log"%(level))
                fh = logging.FileHandler(logFile,'w')
                fh.setLevel(logging.INFO)
                formatter = logging.Formatter('%(asctime)s - %(message)s')
                fh.setFormatter(formatter)
                logger.addHandler(fh)
                logger.setLevel(logging.INFO)
                self.loggers[level] = logger
                
    #@Override
    def ShallLog(self):
        return (self.toConsole or self.toFile)
                
    #@Override         
    def _Log(self,level,msg):
        if(self.toConsole):
            print("%s: %s"%(level,msg))
        if(self.toFile):
            self.loggers[level].info(msg)




'''
 Bjoern Annighoefer 2019
'''
#define the default codec

from .valuecodec import ValueCodec

class NoCodec(ValueCodec):
    def __init__(self):
        pass
    
    def Enc(self,v):
        return v
    
    def Dec(self,c):
        return c
    


import socket
import json
import sys
import pkg_resources
import tornado.ioloop
import tornado.web
from  .eoqserver import PyEoq2WebServer as eoqserver
import signal
import time
import mimetypes # usually imported by tornado, but needed for the mimetypes workaround
import contextlib
from tabulate import tabulate
from functools import partial
from pathlib import Path
from os.path import exists
from yaml import load
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import __main__

__version__ = '0.3.2'

DEFAULT_PATH='./'
DEFAULT_HOST=socket.gethostname()
localpath=str(Path().resolve())
FRAMEWORK_RUN_FLAG=True
apps=[]

xgeedir = pkg_resources.resource_filename(__name__, "core")
xgeelayoutmeta = pkg_resources.resource_filename(__name__, "meta/layout.ecore")

# Fix for XGEE Core Issue #1 Set MIME-Type of static files according to file type
# Ignores system settings for .js and .css since they might be wrongly defined, e.g. as text/plain
mimetypes.add_type('application/javascript', '.js')
mimetypes.add_type('text/css', '.css')

class XGEEConfigHandler(tornado.web.RequestHandler):
    ''' Returns the XGEE configuration as JSON '''
    def initialize(self, xgee_app):
        self.xgee_app=xgee_app
    def get(self):
        xgeeapp_conf='/* Generated XGEE app configuration by XGEE launcher */\n var $XGEE='+str(json.dumps({'app':{'url':self.xgee_app.getWebInstanceUrl()},'eoq':{'url': self.xgee_app.getEoqInstanceUrl()}}))+';'
        self.write(xgeeapp_conf)
    
class EoqConfiguration:
    '''' Default EOQ Configuration '''
    def __init__(self):
        self.ws='/ws/eoq.do'
        self.port=8000
        self.workspaceDir='./workspace'
        self.metaDir='./.meta'
        self.actions=1 # actions are enabled by default
        self.actionsDir='./workspace/actions'
        self.timeout=10.0
        self.backup=0 # backups are disabled by default
        self.backupDir='./backup'
        self.logDir='./log'
        self.logToFile=1 # logging to file is enabled by default
        self.logToConsole=0 # logging to console is disabled by default
        self.autosave=5.0 # the timeout after which autosave is triggered
        self.trackFileChanges=0 # whether tracking of changes on the filesystems is enabled (default:OFF)
        self.maxChanges=10000 # how many changes shall be remembered until the oldest change is forgotten (1 to 1000000)
        self.enableBenchmark=0 # measure time for queries and commands




class XGEEApp:
    def __init__(self, app_path, app_folder_name, app_conf):
        self.app_conf=app_conf
        self.path=app_path
        self.app_folder_name=app_folder_name

        # Prepare Application configuration
        self.name=str(self.get_conf_parameter('app/name'))
        if self.get_conf_parameter('app/name') is None:
            self.name='Unnamed Application'

        self.appHost=str(self.get_conf_parameter('app/host'))
        if self.get_conf_parameter('app/host') is None:
            self.appHost=DEFAULT_HOST       
        self.port=self.get_conf_parameter('app/port')
        self.webInstance=None

        # Prepare EOQ configuration
        self.eoqHost=str(self.get_conf_parameter('eoq/host'))
        if self.get_conf_parameter('eoq/host') is None:
            self.eoqHost=DEFAULT_HOST     
        self.eoqPort=self.get_conf_parameter('eoq/port')
        self.eoqInstance=None
        self.eoqThread=None
        self.eoqConfiguration=None
        if self.eoqPort is not None:
            self.eoqConfiguration=EoqConfiguration()
            self.eoqConfiguration.port=self.eoqPort

            # Configure workspace directory
            if self.get_conf_parameter('eoq/workspace') is not None:
                self.eoqConfiguration.workspaceDir=self.path+str(self.get_conf_parameter('eoq/workspace')).lstrip("./")
            else:
                self.eoqConfiguration.workspaceDir=self.path+'workspace'

            # Configure meta directory
            if self.get_conf_parameter('eoq/meta') is not None:
                self.eoqConfiguration.metaDir=localpath+'/'+self.path+str(self.get_conf_parameter('eoq/meta')).lstrip("./")
            else:
                self.eoqConfiguration.metaDir=localpath+'/'+self.path+'meta'

            # Configure actions directory
            if self.get_conf_parameter('eoq/actions') is not None:
                self.eoqConfiguration.actionsDir=localpath+'/'+self.path+str(self.get_conf_parameter('eoq/actions')).lstrip("./")
            else:
                self.eoqConfiguration.actionsDir=localpath+'/'+self.path+'actions'

            # Configure log directory
            if self.get_conf_parameter('eoq/log') is not None:
                self.eoqConfiguration.logDir=localpath+'/'+self.path+str(self.get_conf_parameter('eoq/log')).lstrip("./")
            else:
                self.eoqConfiguration.logDir=localpath+'/'+self.path+'log'

            # Configure backup directory
            if self.get_conf_parameter('eoq/backup') is not None:
                self.eoqConfiguration.backupDir=localpath+'/'+self.path+str(self.get_conf_parameter('eoq/backup')).lstrip("./")
                self.eoqConfiguration.backup=1 # backups are enabled if a directory was set
            else:
                self.eoqConfiguration.backupDir=localpath+'/'+self.path+'backup'

            # option to log to console
            if self.get_conf_parameter('eoq/consolelog') is not None:
                self.eoqConfiguration.logToConsole=self.get_conf_parameter('eoq/consolelog')

            # Configure autosave
            if self.get_conf_parameter('eoq/autosave') is not None:
                self.eoqConfiguration.autosave=self.get_conf_parameter('eoq/consolelog')

    def getName(self):
        return self.name

    def getPath(self):
        return self.path
    
    def getWebInstanceUrl(self, default='---'):
        ''' Gets the web instance URL '''
        res=default
        if self.webInstance is not None:
            res='http://'+self.appHost+':'+str(self.port)
        return res

    def getEoqInstanceUrl(self,default='---'):
        ''' Gets the eoq instance URL '''
        res=default
        if self.eoqInstance is not None:
            res='ws://'+self.eoqHost+':'+str(self.eoqPort)+'/ws/eoq.do'
        return res

    def launch(self): 
        ''' Launches an XGEE application '''
        webInstance=0
        eoqInstance=0
        if self.port is not None:
            application = tornado.web.Application([
            (r"/xgee-core/(.*)", tornado.web.StaticFileHandler, {"path": str(xgeedir)}),
            (r"/xgee-conf/app.conf.js", XGEEConfigHandler, {"xgee_app":self}),
            (r"/(.*)", tornado.web.StaticFileHandler, {"path": self.path+"/app","default_filename":"index.html"}),
            ])
            self.webInstance=None                           
            self.webInstance = tornado.httpserver.HTTPServer(application)
            self.webInstance.listen(self.port)
            webInstance=1
        else:
            print('App not hosted.')
        if self.eoqPort is not None:
            self.eoqInstance = eoqserver.PyeoqWebSocketServer()
            self.eoqThread=eoqserver.threading.Thread(target=self.eoqInstance.Start, args=(self.eoqConfiguration,))
            self.eoqThread.start()
            while(True):
                if self.eoqInstance.mdbProvider is not None:
                    break
            # Inject meta
            self.eoqInstance.mdbProvider._PyEcoreWorkspaceMdbProvider__LoadMetaModelResource(str(xgeelayoutmeta))
            eoqInstance=1
        else:
            print('EOQ workspace not hosted')
        return webInstance,eoqInstance


    def shutdownWebInstance(self):
        ''' Shuts down the web-instance if applicable'''
        if self.webInstance is not None:
            self.webInstance.stop()

    def shutdownEoqInstance(self):
        ''' Shuts down the web-instance if applicable'''
        if self.eoqInstance is not None:
            self.eoqInstance.Stop(None,None)
        time.sleep(1)
        if self.eoqThread is not None:
            self.eoqThread.join()

    def get_conf_parameter(self,name):
        ''' Retrieves the configuration parameters from the app configuration '''
        res=None
        parameters=name.split("/")
        conf_item=self.app_conf
        for param in parameters:
            if conf_item is None:
                break
            else:
                if param in conf_item:
                    conf_item=conf_item[param]
                else:
                    conf_item=None  
        else:
            res=conf_item        
        return res


def shutdownTornado():
    ''' '''
    io_loop = tornado.ioloop.IOLoop.current()
    io_loop.stop()
    global FRAMEWORK_RUN_FLAG
    for app in apps:
        app.shutdownWebInstance()
    FRAMEWORK_RUN_FLAG=False


def stop(signal,frame):
    ''' Stop XGEE framework '''
    global FRAMEWORK_RUN_FLAG
    print('Shutting down XGEE framework...')
    FRAMEWORK_RUN_FLAG=False    
    tornado.ioloop.IOLoop.current().add_callback_from_signal(shutdownTornado)  

  

 
    


def start(path2apps):    
    ''' Start XGEE framework '''

    global FRAMEWORK_RUN_FLAG
    print("""   
                 _  _  ___  ____  ____ 
                ( \/ )/ __)( ___)( ___)
                 )  (( (_-. )__)  )__) 
                (_/\_)\___/(____)(____)
                VERSION """+str(__version__)+"""

                        """);
                    

    p = Path(path2apps)
    conffile = Path(str(p)+'/configuration.yml')
    if conffile.is_file():
        appname='.'
        app_conf=load(conffile.read_text(),Loader=Loader)
        apps.append(XGEEApp(str(p)+'/'+str(appname)+"/",str(appname),app_conf))
    else:
        # Read app folder and configure XGEE apps   
        if p.is_dir():
            # Instantiate XGEE apps
            appnames=set([f.name for f in p.iterdir() if f.is_dir()])
            for appname in appnames:
                conf = Path(str(p)+'/'+str(appname)+"/configuration.yml")       
                if conf.is_file():
                    app_conf=load(conf.read_text(),Loader=Loader)
                    apps.append(XGEEApp(str(p)+'/'+str(appname)+"/",str(appname),app_conf))
        else:
            print('Invalid directory supplied')

    # Launch XGEE apps
    if len(apps):      
        webInstances=0
        eoqInstances=0
        for app in apps:
            webInstance,eoqInstance=app.launch()
            webInstances+=webInstance
            eoqInstances+=eoqInstance

        print('')
        print('Your XGEE apps are ready:')
        table=[[app.getName(),app.getPath(),app.getWebInstanceUrl(),app.getEoqInstanceUrl()] for app in apps ]
        print(tabulate(table,["App Name","App Path","Web Instance", "EOQ Instance"],tablefmt="fancy_grid"))
        print('')

        if webInstances>0:        
            tornado.ioloop.IOLoop.instance().start()

        if eoqInstances==0:
            FRAMEWORK_RUN_FLAG=False

        # Loop To keep running even if no webinstance is hosted until termination
        while(FRAMEWORK_RUN_FLAG):
            time.sleep(1)

        for app in apps:
            app.shutdownEoqInstance()

        print('XGEE framework stopped.')



          
if __name__ == "__main__":
    signal.signal(signal.SIGTERM, stop)
    signal.signal(signal.SIGINT,  stop)
    path2apps=DEFAULT_PATH
    if len(sys.argv)>1:
        path2apps=sys.argv[1]
    start(path2apps) 

        
    






from .localdomain import LocalDomain
from .multiprocessingqueuedomainhost import MultiprocessingQueueDomainHost
from .multiprocessingqueuedomainclient import MultiprocessingQueueDomainClient
from .simplepythondomainwrapper import SimplePythonDomainWrapper

__version__ = "1.0.0"


'''
Created on 24.03.2019

Wrapper class that enables the use of EOQ domains without 
using commands. Primarily this was written to keep old EOQ examples operational.

@author: Annighoefer
'''

from ..model import ObjectRefValue,CommandA,DomainA
from .. import ValueParser
from .. import QueryParser
from .. import CommandParser
from .. import ResultParser
from ..error import EoqGeneralError

class SimplePythonDomainWrapper():
    def __init__(self,domain : DomainA):
        self.domain = domain
        
    def Do(self,command : CommandA):
        return self.domain.Do(command)

    def Retrieve(self,ptarget,queryStr):
        if(None==ptarget):
            ptarget = ObjectRefValue(v=0)
        target = ValueParser.PythonToValue(ptarget)
        query = QueryParser.StringToQuery(queryStr)
        command = CommandParser.RetrieveCommand(target, query)
        result = self.domain.Do(command)
        if(ResultParser.IsResultNok(result)):
            raise EoqGeneralError(result.code,result.message)
        return ValueParser.ValueToPython(result.value)
    
    def Create(self,packageNsUri,className,n=1):
        command = CommandParser.CreateCommand(packageNsUri, className, n)
        result = self.domain.Do(command)
        if(ResultParser.IsResultNok(result)):
            raise EoqGeneralError(result.code,result.message)
        return ValueParser.ValueToPython(result.value)
    
    def Update(self,ptarget,queryStr,pvalue):
        if(None==ptarget):
            ptarget = ObjectRefValue(v=0)
        target = ValueParser.PythonToValue(ptarget)
        query = QueryParser.StringToQuery(queryStr)
        value = ValueParser.PythonToValue(pvalue)
        command = CommandParser.UpdateCommand(target, query, value)
        result = self.domain.Do(command)
        if(ResultParser.IsResultNok(result)):
            raise EoqGeneralError(result.code,result.message)
        return ValueParser.ValueToPython(result.target)
    
    def Clone(self,ptarget,cloneOp='FULL'):
        if(None==ptarget):
            ptarget = ObjectRefValue(v=0)
        target = ValueParser.PythonToValue(ptarget)
        cmdStr = "CLONE #%d %s"%(target.v,cloneOp)
        cmd = CommandParser.StringToCommand(cmdStr)
        result = self.domain.Do(cmd)
        if(ResultParser.IsResultNok(result)):
            raise EoqGeneralError(result.code,result.message)
        return ValueParser.ValueToPython(result.value)
       
   



'''
 2019 Bjoern Annighoefer
'''
from ..util.error import EoqError

class CallTypes:
    SYN = 'SYN'
    ASY = 'ASY'
    
class CallStatus: 
    INI = 'INI' #initiated
    RUN = 'RUN' #running
    WAI = 'WAI' #waiting (e.g. for user input)
    ABO = 'ABO' #aborted
    ERR = 'ERR' #error
    FIN = 'FIN' #finished (successfully)
 
CALL_STATUS_FINAL_TYPES = [CallStatus.ABO,CallStatus.ERR,CallStatus.FIN] 

class CallOptions:
    def __init__(self):
        self.silent = False
        self.eventsonly = False
        self.timeout = 0.0 #0.0 = no timeout
        self.autoobserve = False
        
    def ToArray(self):
        return [
                'silent',self.silent,
                'eventsonly',self.eventsonly,
                'timeout',self.timeout,
                'autoobserve',self.autoobserve
                ]
        
    def FromArray(self,optionsArr):
        n = len(optionsArr)
        if(n % 2 == 1):
            raise EoqError(0,'Call options array must have an even length, but got length %d.'%(n))
        for i in range(0,n,2):
            name = optionsArr[i]
            val = optionsArr[i+1]
            if('silent' == name):
                if(isinstance(val,bool)):
                    self.silent = val
                else:
                    raise EoqError(0,'Wrong data type for call option %s. Expected bool, but got %s'%(name,val))
            elif('eventsonly' == name):
                if(isinstance(val,bool)):
                    self.eventsonly = val
                else:
                    raise EoqError(0,'Wrong data type for call option %s. Expected bool, but got %s'%(name,val))
            elif('timeout' == name):
                if(isinstance(val,float)):
                    self.timeout = val
                else:
                    raise EoqError(0,'Wrong data type for call option %s. Expected float, but got %s'%(name,val))
            elif('autoobserve' == name):
                if(isinstance(val,bool)):
                    self.autoobserve = val
                else:
                    raise EoqError(0,'Wrong data type for call option %s. Expected bool, but got %s'%(name,val))
            else:
                raise EoqError(0,'Unknown call option %s.'%(name))
                
                

'''
 CALL 
''' 
class Call: 
    def __init__(self,callId,domain,name,args,opts,handler,ctype,sessionId,status=CallStatus.INI,parentSessionId=None):
        self.callId = callId
        self.domain = domain
        self.name = name
        self.args = args
        self.opts = opts
        self.handler = handler
        self.ctype = ctype
        self.status = status
        self.sessionId = sessionId
        self.parentSessionId = parentSessionId
        self.outputs= {}
        self.hasValue = False
        self.value = None
        self.startTime = None
        self.endTime = None
        
        


from pyecore.resources import global_registry
from .model import getEClassifier, eClassifiers
from .model import name, nsURI, nsPrefix, eClass
from .model import NamedElementA, DomainA, Metamodel, MetaModelSourcesE, CommandInfo, Transaction, Change, ActionInfo, ActionParameter, Choice, ActionHandlerA, CommandParameter, LocalDomainA, Directory, FileResourceA, FileResourceTypesE, ModelResource, TextResource, RawResource, Query, Segment, SegmentTypesE, ResourceSegment, PathSegment, ClassSegment, InstanceOfSegment, IdSegment, MetaSegment, MetaSegmentIdentifiersE, HistorySegment, Depth, Index, IndexTypesE, NumberIndex, RangeIndex, AddIndex, RemoveIndex, FlattenIndex, Selector, Operator, OperatorTypesE, EqualOperator, NotEqualOperator, GreaterOperator, LessOperator, Value, ValueTypesE, IntValue, FloatValue, BoolValue, StringValue, ObjectRefValue, EmptyValue, ListValue, HistoryRefValue, OperationValue, ObjectValue, SourceClass, ReturnMultiplicity, ReturnMultiplicityTypeE, UnchangedReturnMuliplicity, FlattenedReturnMultiplicity, ForceSingleReturnMultiplicity, CommandA, CommandTypesE, HelloCommand, GoodbyeCommand, SessionCommand, StatusCommand, ChangesCommand, RetrieveCommand, CreateCommand, UpdateCommand, CloneModesE, CloneCommand, CallCommand, AsyncCallCommand, CallStatusCommand, AbortCallCommand, CompoundCommand, ResultA, ResultTypesE, HelloResult, GoodbyeResult, SessionResult, StatusResult, ChangesResult, RetrieveResult, CreateResult, UpdateResult, CloneResult, CallResult, AsyncCallResult, CallStatusResult, AbortCallResult, ErrorResult, CompoundResult, CallStatusE, Object, ActionCall, CallChannel, CallChannelData, CallChannelTypesE, QueryValue, ListOpSegment, ListOpSegmentTypesE, RecursiveSegmetA, LinearSegmetA, SegmentProcessingTypesE, SelectorSegment, IndexSegment, PathElementA

from pyecore.ecore import EObject

from . import model

__all__ = ['NamedElementA', 'DomainA', 'Metamodel', 'MetaModelSourcesE', 'CommandInfo', 'Transaction', 'Change', 'ActionInfo', 'ActionParameter', 'Choice', 'ActionHandlerA', 'CommandParameter', 'LocalDomainA', 'Directory', 'FileResourceA', 'FileResourceTypesE', 'ModelResource', 'TextResource', 'RawResource', 'Query', 'Segment', 'SegmentTypesE', 'ResourceSegment', 'PathSegment', 'ClassSegment', 'InstanceOfSegment', 'IdSegment', 'MetaSegment', 'MetaSegmentIdentifiersE', 'HistorySegment', 'Depth', 'Index', 'IndexTypesE', 'NumberIndex', 'RangeIndex', 'AddIndex', 'RemoveIndex', 'FlattenIndex', 'Selector', 'Operator', 'OperatorTypesE', 'EqualOperator', 'NotEqualOperator', 'GreaterOperator', 'LessOperator', 'Value', 'ValueTypesE', 'IntValue', 'FloatValue', 'BoolValue', 'StringValue', 'ObjectRefValue', 'EmptyValue', 'ListValue', 'HistoryRefValue', 'OperationValue', 'ObjectValue', 'SourceClass', 'ReturnMultiplicity', 'ReturnMultiplicityTypeE',
           'UnchangedReturnMuliplicity', 'FlattenedReturnMultiplicity', 'ForceSingleReturnMultiplicity', 'CommandA', 'CommandTypesE', 'HelloCommand', 'GoodbyeCommand', 'SessionCommand', 'StatusCommand', 'ChangesCommand', 'RetrieveCommand', 'CreateCommand', 'UpdateCommand', 'CloneModesE', 'CloneCommand', 'CallCommand', 'AsyncCallCommand', 'CallStatusCommand', 'AbortCallCommand', 'CompoundCommand', 'ResultA', 'ResultTypesE', 'HelloResult', 'GoodbyeResult', 'SessionResult', 'StatusResult', 'ChangesResult', 'RetrieveResult', 'CreateResult', 'UpdateResult', 'CloneResult', 'CallResult', 'AsyncCallResult', 'CallStatusResult', 'AbortCallResult', 'ErrorResult', 'CompoundResult', 'CallStatusE', 'Object', 'ActionCall', 'CallChannel', 'CallChannelData', 'CallChannelTypesE', 'QueryValue', 'ListOpSegment', 'ListOpSegmentTypesE', 'RecursiveSegmetA', 'LinearSegmetA', 'SegmentProcessingTypesE', 'SelectorSegment', 'IndexSegment', 'PathElementA']

eSubpackages = []
eSuperPackage = None
model.eSubpackages = eSubpackages
model.eSuperPackage = eSuperPackage

DomainA.metamodels.eType = Metamodel
DomainA.transactions.eType = Transaction
DomainA.commands.eType = CommandInfo
DomainA.models.eType = EObject
DomainA.actions.eType = ActionInfo
DomainA.currentTransaction.eType = Transaction
DomainA.changes.eType = Change
DomainA.actionCalls.eType = ActionCall
Metamodel.package.eType = EObject
CommandInfo.parameters.eType = CommandParameter
CommandInfo.results.eType = CommandParameter
Transaction.history.eType = Value
Transaction.changes.eType = Change
Change.target.eType = Value
Change.query.eType = Query
Change.newValue.eType = Value
Change.oldValue.eType = Value
ActionInfo.parameters.eType = ActionParameter
ActionInfo.results.eType = ActionParameter
ActionInfo.handler.eType = ActionHandlerA
ActionParameter.choices.eType = Choice
Directory.subdirectories.eType = Directory
Directory.resources.eType = FileResourceA
ModelResource.contents.eType = EObject
ModelResource.domain.eType = LocalDomainA
Query.segments.eType = Segment
Query.sourceClass.eType = SourceClass
Query.returnMultiplicity.eType = ReturnMultiplicity
Segment.selector.eType = Selector
Segment.index.eType = Index
Segment.depth.eType = Depth
PathSegment.feature.eType = EObject
ClassSegment.clazz.eType = EObject
InstanceOfSegment.clazz.eType = EObject
IdSegment.element.eType = EObject
HistorySegment.element.eType = EObject
Selector.value.eType = Value
Selector.operator.eType = Operator
ListValue.v.eType = Value
ObjectValue.v.eType = EObject
SourceClass.type.eType = EObject
HelloCommand.identification.eType = Value
HelloCommand.sessionId.eType = Value
GoodbyeCommand.sessionId.eType = Value
SessionCommand.sessionId.eType = Value
ChangesCommand.earliestChangeId.eType = Value
RetrieveCommand.target.eType = Value
RetrieveCommand.query.eType = Query
CreateCommand.packageNsUri.eType = Value
CreateCommand.className.eType = Value
CreateCommand.n.eType = Value
UpdateCommand.target.eType = Value
UpdateCommand.query.eType = Query
UpdateCommand.value.eType = Value
CloneCommand.target.eType = Value
CallCommand.action.eType = Value
CallCommand.args.eType = ListValue
AsyncCallCommand.action.eType = Value
AsyncCallCommand.args.eType = ListValue
CallStatusCommand.callId.eType = Value
AbortCallCommand.callId.eType = Value
CompoundCommand.commands.eType = CommandA
ChangesResult.changes.eType = ListValue
RetrieveResult.value.eType = Value
CreateResult.value.eType = Value
UpdateResult.target.eType = Value
CloneResult.value.eType = Value
CallResult.returnValues.eType = Value
CallStatusResult.result.eType = Value
CompoundResult.results.eType = ResultA
ActionCall.channels.eType = CallChannel
ActionCall.args.eType = Value
ActionCall.handler.eType = ActionHandlerA
ActionCall.returnValues.eType = Value
CallChannel.data.eType = CallChannelData
QueryValue.v.eType = Query

otherClassifiers = [MetaModelSourcesE, FileResourceTypesE, SegmentTypesE, MetaSegmentIdentifiersE, IndexTypesE, OperatorTypesE, ValueTypesE,
                    ReturnMultiplicityTypeE, CommandTypesE, CloneModesE, ResultTypesE, CallStatusE, Object, CallChannelTypesE, ListOpSegmentTypesE, SegmentProcessingTypesE]

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [model] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack
